from __future__ import print_function, unicode_literals

import uuid
from itertools import chain
from numbers import Number
from antelope import CatalogRef, BaseEntity, PropertyExists

from synonym_dict import LowerDict


entity_types = ('process', 'flow', 'quantity', 'fragment')
entity_refs = {
    'process': 'exchange',
    'flow': 'quantity',
    'quantity': 'unit',
    'fragment': 'fragment'
}


def concatenate(*lists):
    return chain(*lists)


class EntityInitializationError(Exception):
    pass


class EntityMergeError(Exception):
    pass


class LcEntity(BaseEntity):
    """
    All LC entities behave like dicts, but they all have some common properties, defined here.
    """
    _pre_fields = ['Name']
    _new_fields = []
    _ref_field = ''
    _post_fields = ['Comment']

    _origin = None

    def __init__(self, entity_type, external_ref, origin=None, entity_uuid=None, **kwargs):

        if external_ref is None:
            if entity_uuid is None:
                raise EntityInitializationError('At least one of entity_uuid, external_ref must be provided')
            external_ref = str(entity_uuid)

        self._external_ref = str(external_ref)

        self._uuid = None

        if entity_uuid is not None:
            self.uuid = entity_uuid

        self._d = LowerDict()

        self._entity_type = entity_type
        self._reference_entity = None

        if origin is not None:
            self.origin = origin

        self._d['Name'] = self._external_ref
        self._d['Comment'] = ''

        self._query_ref = None  # memoize this

        for k, v in kwargs.items():
            if v is None:
                continue
            self[k] = v

    @property
    def reference_entity(self):
        return self._reference_entity

    def make_ref(self, query):
        if self._query_ref is None:
            d = dict()
            for k in self.signature_fields():
                if k == self._ref_field:
                    continue
                if k in self._d:
                    d[k] = self._d[k]
            self._query_ref = CatalogRef.from_query(self.external_ref, query, self.entity_type,
                                                    uuid=self.uuid, **d)
        return self._query_ref

    @property
    def entity_type(self):
        return self._entity_type

    @property
    def origin(self):
        return self._origin

    @property
    def is_entity(self):
        """
        Used to distinguish between entities and catalog refs (which answer False)
        :return: True for LcEntity subclasses
        """
        return True

    def map_origin(self, omap, fallback=None):
        """
        This is used to propagate a change in origin semantics. Provide a dict that maps old origins to new origins.
        External ref should remain the same with respect to the new origin.
        :param omap: dict mapping old origin to new origin
        :param fallback: if present, use in cases where old origin not found
        :return:
        """
        if self._origin in omap:
            self._origin = omap[self._origin]
        elif fallback is not None:
            self._origin = fallback

    @origin.setter
    def origin(self, value):
        if self._origin is None:
            self._origin = value
        else:
            raise PropertyExists('Origin already set to %s' % self._origin)

    def signature_fields(self):
        return concatenate(self._pre_fields, self._new_fields,
                           [self._ref_field] if self._ref_field is not [] else [], self._post_fields)

    @property
    def reference_field(self):
        return self._ref_field

    @property
    def external_ref(self):
        return self._external_ref

    def get_signature(self):
        k = dict()
        for i in self.signature_fields():
            k[i] = self[i]
        return k

    @property
    def uuid(self):
        return self._uuid

    @uuid.setter
    def uuid(self, key):
        if self._uuid is not None:
            raise PropertyExists('UUID has already been specified! %s' % self._uuid)
        if isinstance(key, uuid.UUID):
            self._uuid = str(key)
        else:
            self._uuid = str(uuid.UUID(key))

    def _validate_reference(self, ref_entity):
        if ref_entity is None:
            # raise ValueError('Null reference')
            return False  # allow none references
        if ref_entity.entity_type != entity_refs[self.entity_type]:
            raise TypeError("Type Mismatch on reference entity: expected %s, found %s" % (entity_refs[self.entity_type],
                                                                                          ref_entity.entity_type))
        return True

    def _set_reference(self, ref_entity):
        """
        set the entity's reference value.  Can be overridden
        :param ref_entity:
        :return:
        """
        self._validate_reference(ref_entity)
        self._reference_entity = ref_entity

    def has_property(self, prop):
        return prop in self._d

    def properties(self):
        for i in self._d.keys():
            yield i

    def get_properties(self):
        """
        dict of properties and values for a given entity
        :return:
        """
        d = dict()
        for i in self.properties():
            d[i] = self._d[i]
        return d

    def update(self, d):
        self._d.update(d)

    def validate(self):
        valid = True
        if self.reference_entity is not None:
            try:
                self._validate_reference(self.reference_entity)
            except TypeError:
                print("Reference entity type %s is wrong for %s (%s)" %
                      (self.reference_entity.entity_type,
                       self.entity_type,
                       entity_refs[self.entity_type]))
                valid = False
        for i in self.signature_fields():
            try:
                self[i]
            except KeyError:
                print("Required field %s does not exist" % i)
                valid = False
        return valid

    def _print_ref_field(self):
        if self.reference_entity is None:
            return None
        else:
            return '%s' % self.reference_entity.external_ref

    def serialize(self, domesticate=False, drop_fields=()):
        j = {
            'entityType': self.entity_type,
            'externalId': self.external_ref,
            'origin': self.origin,
            self._ref_field: self._print_ref_field(),
        }
        if domesticate or self._origin is None:
            j.pop('origin')
        for k, v in self._d.items():
            if k in drop_fields:
                continue
            if v is None:
                continue
            elif isinstance(v, list):
                j[k] = v
            elif isinstance(v, set):
                j[k] = sorted(list(v))
            elif isinstance(v, Number):
                j[k] = v
            elif isinstance(v, bool):
                j[k] = v
            elif isinstance(v, LcEntity):
                j[k] = {"origin": v.origin,
                        "externalId": v.external_ref,
                        "entity_type": v.entity_type}
            elif isinstance(v, dict):
                j[k] = v
            else:
                j[k] = str(v)
        return j

    def __getitem__(self, item):
        if item.lower() == self._ref_field.lower():
            return self.reference_entity
        elif item == 'EntityType':
            return self.entity_type
        else:
            # don't catch KeyErrors here-- leave that to subclasses
            return self._d[item]

    def get(self, item, default=None):
        try:
            return self.__getitem__(item)
        except KeyError:
            return default

    def __setitem__(self, key, value):
        if key.lower() in (self._ref_field.lower(), 'reference', 'referenceentity', 'reference_entity'):
            self._set_reference(value)
        elif key.lower() in ('entityid', 'entitytype', 'externalid', 'origin'):
            raise KeyError('Disallowed Keyname %s' % key)
        else:
            self._d[key] = value

    def merge(self, other):
        if False:  # not isinstance(other, LcEntity):  ## This is not a requirement! cf. EntityRefs, Disclosure objs
            raise EntityMergeError('Incoming is not an LcEntity: %s' % other)
        elif self.entity_type != other.entity_type:
            raise EntityMergeError('Incoming entity type %s mismatch with %s' % (other.entity_type, self.entity_type))
        elif self.external_ref != other.external_ref:
            raise EntityMergeError('Incoming External ref %s conflicts with existing %s' % (other.external_ref,
                                                                                            self.external_ref))
        else:
            # if self.origin != other.origin:
            #     print('Merging entities with differing origin: \nnew: %s\nexisting: %s'% (other.origin, self.origin))
            for k in other.properties():
                if k not in self._d.keys():
                    print('Merge: Adding key %s: %s' % (k, other[k]))
                    self[k] = other[k]

    def show(self):
        print('%s Entity (ref %s)' % (self.entity_type.title(), self.external_ref))
        print('origin: %s' % self.origin)
        if self.entity_type == 'process':
            for i in self.reference_entity:
                print('reference: %s' % i)
        else:
            print('reference: %s' % self.reference_entity)
        fix = ['Name', 'Comment']
        postfix = set(str(k) for k in self._d.keys()).difference(fix)
        ml = len(max(self._d.keys(), key=len))
        for k in fix:
            print('%*s: %s' % (ml, k, self._d[k]))
        for k in postfix:
            print('%*s: %s' % (ml, k, self._d[k]))

    def __str__(self):
        return 'LC %s: %s' % (self.entity_type, self._d['Name'])

    @property
    def _name(self):
        return str(self)

    def __hash__(self):
        """
        External ref is set by the end of __init__ and is immutable (except for fragments-- which use uuid for hash)
        :return:
        """
        if self._origin is None:
            raise AttributeError('Origin not set!')
        return hash(self.link)

    def __eq__(self, other):
        """
        two entities are equal if their types, origins, and external references are the same.
        internal refs do not need to be equal; reference entities do not need to be equal
        :return:
        """
        if other is None:
            return False
        # if not isinstance(other, LcEntity):  # taking this out so that CatalogRefs and entities can be compared
        #     return False
        try:
            is_eq = (self.external_ref == other.external_ref
                     and self.origin == other.origin
                     and self.entity_type == other.entity_type)
        except AttributeError:
            is_eq = False
        return is_eq



import unittest

from ..archives import LcArchive
from ..exchanges import Exchange, ExchangeValue
from ..exchanges import AmbiguousReferenceError, DuplicateExchangeError, ExchangeError
from ..entities.tests import refinery_archive

grid_id = '96bffbb9-b875-36cf-8a11-5723c9d239d9'
petro_id = '0aaf1e13-5d80-37f9-b7bb-81a6b8965c71'
nox_id = '1827a862-ccac-37ac-9ad7-d2dffe71058c'
ocean_id = '455470fa-1a78-344b-ae09-ffa4e22c1f8b'
ng_id = '6cb3e4a8-d7e0-3566-ae66-8022182af846'
diesel_id = 'd939590b-a0d7-310c-8952-9921ed64a078'
resid_id = '562918fe-3ff5-33f4-abaa-d8c615380d25'


class ExchangesTest(unittest.TestCase):
    """
    An exchange requires a process, flow, direction, and optional termination specifier.
    The only things really to test ore setting and querying exchange values

    Expected behavior:
    __setitem__(key, value):
       * key.process must be the same as self.process
       * key must be a reference exchange
       * exchange must not already have an allocation factor specified for the given key
       * if key is self, unallocated value is set (allocated value of reference exchange is always 1 or 0)
       * if self is reference exchange but not key, value must be 0 (and setting it has no effect)
       notes:
       = if process is using allocation by quantity, the specified value will never be returned

    __getitem__(key):
       * if self is a reference item, the allocated exchange value must be 1 if key is self; 0 otherwise
       * if key is a reference item:
         - if process is using allocation by quantity, compute the NORMALIZED factor and return it
         - elif key is in the exchange's value dict, return stored value
         - else return 0
       * else normalize unallocated value by key's value and return it

    """
    @classmethod
    def setUpClass(cls):
        cls.A = LcArchive.from_file(refinery_archive)
        cls.grid = cls.A[grid_id]
        cls.petro = cls.A[petro_id]
        cls.diesel = cls.A[diesel_id]
        cls.petro_nox = next(cls.petro.exchange_values(cls.A[nox_id]))  # pull the NOx exchange
        cls.petro_ocean = next(cls.petro.exchange_values(cls.A[ocean_id]))  # ocean freight exchange
        cls.petro_d = cls.petro.reference(cls.diesel)
        cls.petro_r = cls.petro.reference(cls.A[resid_id])
        cls.grid_ng = next(cls.grid.exchange_values(cls.A[ng_id]))  # electricity from natural gas

    def test_equality(self):
        ea = Exchange(self.petro, self.diesel, direction='Output')
        eb = Exchange(self.petro, self.diesel, direction='Output', termination='1234567')
        self.assertNotEqual(ea, eb)

    def test_hashing(self):
        ea = Exchange(self.grid, self.diesel, 'Input')
        ev = ExchangeValue(self.grid, self.diesel, 'Input', value=17)
        self.assertEqual(ea, ev)
        k = {ev}
        self.assertIn(ea, k)

    def test_exchange_value(self):
        self.assertAlmostEqual(self.grid.reference().value, 3.6, places=12)

    def test_same_process(self):
        with self.assertRaises(ExchangeError):
            val = self.petro_nox[self.grid.reference()]

        with self.assertRaises(ExchangeError):
            self.petro_nox[self.grid.reference()] = 47

    def test_reference_exchange(self):
        with self.assertRaises(AmbiguousReferenceError):
            self.petro_nox[self.petro_ocean] = 123

        self.assertAlmostEqual(self.petro_d.value, 0.000252345, places=6)
        self.assertIsNotNone(self.petro_nox[self.petro_d])

    def test_reference_alloc(self):
        self.assertEqual(self.petro_d[self.petro_d], 1.0)
        self.assertEqual(self.petro_r[self.petro_r], 1.0)
        self.assertEqual(self.petro_d[self.petro_r], 0.0)

    def test_duplicate_exchange(self):
        with self.assertRaises(DuplicateExchangeError):
            self.petro_d[self.petro_d] = 47

    def test_norm_computation(self):
        self.assertAlmostEqual(self.grid_ng[self.grid.reference()], self.grid_ng.value / self.grid.reference().value,
                               places=10)

    def test_alloc_computation(self):
        # let's spell the whole thing out
        exch = self.petro_ocean
        ref = self.petro_d
        total_req = exch.value  # 4.408 t*km
        total_output = self.petro.alloc_total  # 1.0014 kg
        opt_per_kg = total_req/total_output  # 4.403 t*km / kg

        kg_per_unit_ref = self.petro.alloc_qty.cf(self.petro_d.flow)  # 1 unit diesel = 850 kg

        self.assertAlmostEqual(exch[ref], opt_per_kg * kg_per_unit_ref, places=10)

    def test_alloc_consistency(self):
        """
        For each non-reference item in the inventory, test that the values allocated to each reference, weighted by the
        reference values, sum up to the un-allocated value.
        :return:
        """
        '''
        for k in self.petro.inventory():
            if k.is_reference:
                continue
            total = sum(k[rx] * rx.value for rx in self.petro.references())
            self.assertAlmostEqual(total, k.value, places=10)
        '''
        self.assertTrue(self.petro.test_allocation_consistency(display=False))

    def test_alloc_equality(self):
        d_cf = self.petro.alloc_qty.cf(self.petro_d.flow)
        r_cf = self.petro.alloc_qty.cf(self.petro_r.flow)
        # d_cf = self.petro_d.flow.cf(self.petro.alloc_qty)
        # r_cf = self.petro_r.flow.cf(self.petro.alloc_qty)
        for k in self.petro.inventory():
            if k.is_reference:
                continue
            self.assertAlmostEqual(k[self.petro_d] / d_cf, k[self.petro_r] / r_cf, places=10)


if __name__ == '__main__':
    unittest.main()



from antelope import IndexRequired
from .catalog import StaticCatalog
from ..archives import REF_QTYS, archive_from_json
from ..lc_resource import LcResource, download_file
from ..lcia_engine import DEFAULT_CONTEXTS, DEFAULT_FLOWABLES

from shutil import copy2, rmtree
import os

TEST_ROOT = os.path.join(os.path.dirname(__file__), 'cat-test')  # volatile, inspectable

class LcCatalog(StaticCatalog):
    """
    A catalog that supports adding and manipulating resources during runtime
    """
    def download_file(self, url=None, md5sum=None, force=False, localize=True):
        """
        Download a file from a remote location into the catalog and return its local path.  Optionally validate the
        download with an MD5 digest.
        :param url:
        :param md5sum:
        :param force:
        :param localize: whether to return the filename relative to the catalog root
        :return:
        """
        local_file = os.path.join(self._download_dir, self._source_hash_file(url))
        if os.path.exists(local_file):
            if force:
                print('File exists.. re-downloading.')
            else:
                print('File already downloaded.  Force=True to re-download.')
                if localize:
                    return self._localize_source(local_file)
                return local_file

        download_file(url, local_file, md5sum)

        import magic
        if magic.from_file(local_file).startswith('Microsoft Excel 20'):
            new_file = local_file + '.xlsx'
            os.rename(local_file, new_file)  # openpyxl refuses to open files without an extension
            local_file = new_file

        if localize:
            return self._localize_source(local_file)
        return local_file

    @classmethod
    def make_tester(cls, **kwargs):
        rmtree(TEST_ROOT, ignore_errors=True)
        return cls(TEST_ROOT, **kwargs)

    @classmethod
    def load_tester(cls):
        return cls(TEST_ROOT)

    @property
    def _dirs(self):
        for x in (self._cache_dir, self._index_dir, self.resource_dir, self.archive_dir, self._download_dir):
            yield x

    def _make_rootdir(self):
        for x in self._dirs:
            os.makedirs(x, exist_ok=True)
        if not os.path.exists(self._contexts):
            copy2(DEFAULT_CONTEXTS, self._contexts)
        if not os.path.exists(self._flowables):
            copy2(DEFAULT_FLOWABLES, self._flowables)
        if not os.path.exists(self._reference_qtys):
            copy2(REF_QTYS, self._reference_qtys)

    def __init__(self, rootdir, **kwargs):
        self._rootdir = os.path.abspath(rootdir)
        self._make_rootdir()  # this will be a git clone / fork;; clones reference quantities
        super(LcCatalog, self).__init__(self._rootdir, **kwargs)

    def save_local_changes(self):
        self._qdb.write_to_file(self._reference_qtys, characterizations=True, values=True)
        self.lcia_engine.save_flowables(self._flowables)
        self.lcia_engine.save_contexts(self._contexts)

    def restore_qdb(self, really=False):
        if really:
            copy2(REF_QTYS, self._reference_qtys)
            print('Reference quantities restored. Please re-initialize the catalog.')

    '''
    Create + Add data resources
    '''

    def new_resource(self, reference, source, ds_type, store=True, **kwargs):
        """
        Create a new data resource by specifying its properties directly to the constructor
        :param reference:
        :param source:
        :param ds_type:
        :param store: [True] permanently store this resource
        :param kwargs: interfaces=None, priority=0, static=False; **kwargs passed to archive constructor
        :return:
        """
        source = self._localize_source(source)
        return self._resolver.new_resource(reference, source, ds_type, store=store, **kwargs)  # explicit store= for doc purposes

    def add_resource(self, resource, store=True):
        """
        Add an existing LcResource to the catalog.
        :param resource:
        :param store: [True] permanently store this resource
        :return:
        """
        self._resolver.add_resource(resource, store=store)
        # self._ensure_resource(resource)

    def delete_resource(self, resource, delete_source=None, delete_cache=True):
        """
        Removes the resource from the resolver and also removes the serialization of the resource. Also deletes the
        resource's source under the following circumstances:
         (resource is internal AND resources_with_source(resource.source) is empty AND resource.source is a file)
        This can be overridden using he delete_source param (see below)
        :param resource: an LcResource
        :param delete_source: [None] If None, follow default behavior. If True, delete the source even if it is
         not internal (source will not be deleted if other resources refer to it OR if it is not a file). If False,
         do not delete the source.
        :param delete_cache: [True] whether to delete cache files (you could keep them around if you expect to need
         them again and you don't think the contents will have changed)
        :return:
        """
        self._resolver.delete_resource(resource)
        abs_src = self.abs_path(resource.source)

        if delete_source is False or resource.source is None or not os.path.isfile(abs_src):
            return
        if len([t for t in self._resolver.resources_with_source(resource.source)]) > 0:
            return
        if resource.internal or delete_source:
            if os.path.isdir(abs_src):
                rmtree(abs_src)
            else:
                os.remove(abs_src)
        if delete_cache:
            if os.path.exists(self.cache_file(resource.source)):
                os.remove(self.cache_file(resource.source))
            if os.path.exists(self.cache_file(abs_src)):
                os.remove(self.cache_file(abs_src))

    def add_existing_archive(self, archive, interfaces=None, store=True, **kwargs):
        """
        Makes a resource record out of an existing archive.  by default, saves it in the catalog's resource dir
        :param archive:
        :param interfaces:
        :param store: [True] if False, don't save the record - use it for this session only
        :param kwargs:
        :return:
        """
        res = LcResource.from_archive(archive, interfaces, source=self._localize_source(archive.source), **kwargs)
        self._resolver.add_resource(res, store=store)

    '''
    Manage resources locally
     - index
     - cache
     - static archive (performs load_all())
    '''

    def _index_source(self, source, priority, force=False):
        """
        Instructs the resource to create an index of itself in the specified file; creates a new resource for the
        index
        :param source:
        :param priority:
        :param force:
        :return:
        """
        res = next(r for r in self._resolver.resources_with_source(source))
        res.check(self)
        priority = min([priority, res.priority])  # why are we doing this?? we want index to have higher priority i.e. get loaded second
        stored = self._resolver.is_permanent(res)

        # save configuration hints in derived index
        cfg = None
        if stored:
            if len(res.config['hints']) > 0:
                cfg = {'hints': res.config['hints']}

        inx_file = self._index_file(source)
        inx_local = self._localize_source(inx_file)

        if os.path.exists(inx_file):
            if not force:
                print('Not overwriting existing index. force=True to override.')
                try:
                    ex_res = next(r for r in self._resolver.resources_with_source(inx_local))
                    return ex_res.reference
                except StopIteration:
                    # index file exists, but no matching resource
                    inx = archive_from_json(inx_file)
                    self.new_resource(inx.ref, inx_local, 'json', priority=priority, store=stored,
                                      interfaces='index', _internal=True, static=True, preload_archive=inx,
                                      config=cfg)

                    return inx.ref

            print('Re-indexing %s' % source)
            # TODO: need to delete the old index resource!!
            stale_res = list(self._resolver.resources_with_source(inx_local))
            for stale in stale_res:
                # this should be postponed to after creation of new, but that fails in case of naming collision (bc YYYYMMDD)
                # so golly gee we just delete-first.
                print('deleting %s' % stale.reference)
                self._resolver.delete_resource(stale)

        the_index = res.make_index(inx_file, force=force)
        self.new_resource(the_index.ref, inx_local, 'json', priority=priority, store=stored, interfaces='index',
                          _internal=True, static=True, preload_archive=the_index, config=cfg)

        return the_index.ref

    def index_ref(self, origin, interface=None, source=None, priority=60, force=False, strict=True):
        """
        Creates an index for the specified resource.  'origin' and 'interface' must resolve to one or more LcResources
        that all have the same source specification.  That source archive gets indexed, and index resources are created
        for all the LcResources that were returned.

        Performs load_all() on the source archive, writes the archive to a compressed json file in the local index
        directory, and creates a new LcResource pointing to the JSON file.   Aborts if the index file already exists
        (override with force=True).
        :param origin:
        :param interface: [None]
        :param source: find_single_source input
        :param priority: [60] priority setting for the new index
        :param force: [False] if True, overwrite existing index
        :param strict: [True] whether to be strict
        :return:
        """
        try:
            ix = next(self.gen_interfaces(origin, itype='index', strict=False))
            return ix.origin
        except StopIteration:
            source = self._find_single_source(origin, interface, source=source, strict=strict)
            return self._index_source(source, priority, force=force)

    def cache_ref(self, origin, interface=None, source=None, static=False):
        source = self._find_single_source(origin, interface, source=source)
        self.create_source_cache(source, static=static)

    def create_source_cache(self, source, static=False):
        """
        Creates a cache of the named source's current contents, to speed up access to commonly used entities.
        source must be either a key present in self.sources, or a name or nickname found in self.names
        :param source:
        :param static: [False] create archives of a static archive (use to force archival of a complete database)
        :return:
        """
        res = next(r for r in self._resolver.resources_with_source(source))
        if res.static:
            if not static:
                print('Not archiving static resource %s' % res)
                return
            print('Archiving static resource %s' % res)
        res.check(self)
        res.make_cache(self.cache_file(self._localize_source(source)))

    def _background_for_origin(self, ref, strict=False):
        res = self.get_resource(ref, iface='exchange')
        inx_ref = self.index_ref(ref, interface='exchange', strict=strict)
        bk_file = self._localize_source(os.path.join(self.archive_dir, '%s_background.mat' % inx_ref))
        bk = LcResource(inx_ref, bk_file, 'Background', interfaces='background', priority=99,
                        save_after=True, _internal=True)
        bk.config = res.config
        bk.check(self)  # ImportError if antelope_background pkg not found;; also applies configs
        self.add_resource(bk)
        return bk.make_interface('background')  # when the interface is returned, it will trigger setup_bm

    def gen_interfaces(self, origin, itype=None, strict=False, ):
        """
        Override parent method to also create local backgrounds
        :param origin:
        :param itype:
        :param strict:
        :return:
        """
        for k in super(LcCatalog, self).gen_interfaces(origin, itype=itype, strict=strict):
            yield k

        if itype == 'background':
            if origin.startswith('local') or origin.startswith('test'):
                yield self._background_for_origin(origin, strict=strict)

    def create_descendant(self, origin, interface=None, source=None, force=False, signifier=None, strict=True,
                          priority=None, **kwargs):
        """

        :param origin:
        :param interface:
        :param source:
        :param force: overwrite if exists
        :param signifier: semantic descriptor for the new descendant (optional)
        :param strict:
        :param priority:
        :param kwargs:
        :return:
        """
        res = self.get_resource(origin, iface=interface, source=source, strict=strict)
        new_ref = res.archive.create_descendant(self.archive_dir, signifier=signifier, force=force)
        print('Created archive with reference %s' % new_ref)
        ar = res.archive
        prio = priority or res.priority
        self.add_existing_archive(ar, interfaces=res.interfaces, priority=prio, **kwargs)
        res.remove_archive()



from collections import namedtuple
from antelope import ConfigureInterface, check_direction

from .basic import BasicImplementation


ValidConfig = namedtuple('ValidConfig', ('nargs', 'argtypes'))

valid_configs = {  # 0_ indicates the arg is allowed to be None
    'hints': ValidConfig(3, ('hint', 'str', 'str')),
    'set_reference': ValidConfig(3, ('process', 'flow', '0_direction')),
    'unset_reference': ValidConfig(3, ('0_process', '0_flow', '0_direction')),
    'characterize_flow': ValidConfig(3, ('flow', 'quantity', 'float')),
    'allocate_by_quantity': ValidConfig(2, ('process', 'quantity')),
    'prefer_provider': ValidConfig(2, ('flow', '0_process'))
}


class CoreConfigureImplementation(BasicImplementation, ConfigureInterface):

    _config_options = ()

    def _apply_config(self, config, option, **kwargs):
        if option in config:
            for k in config[option]:
                getattr(self, option)(*k, **kwargs)

    def apply_config(self, config, **kwargs):
        """
        Apply a collection of configuration objects to the archive.

        All options should be supplied as a dict of iterables, where the keys are the configuration methods and
        the entries in each iterable are tuples corresponding to the configuration parameters, properly ordered.

        Configuration options should be idempotent; applying them several times should have the same effect as
        applying them once.
        :return:
        """
        print('Applying configuration to %s' % self._archive)
        # re_index = False
        '''
        # self._apply_config(config, 'add_terms')
        for k in _config['add_terms']:
            print('adding %s synonyms (%s|%d)' % (k[0], k[1], len(k) - 1))
            self.add_terms(*k)
        '''
        for opt in self._config_options:
            self._apply_config(config, opt, **kwargs)

        if hasattr(self._archive, 'ti'):
            self._archive.make_interface('index').re_index()

    def check_config(self, config, c_args, **kwargs):
        """
        Uses the config schema valid_configs to validate config arguments.  A leading '0_' indicates that the argument
        is permitted to be None.

        :param config: a configuration entry
        :param c_args: a tuple of args
        :param kwargs:
        :return:
        """
        vc = valid_configs[config]
        if len(c_args) != vc.nargs:
            raise ValueError('Wrong number of arguments (%d supplied; %d required)' % (len(c_args), vc.nargs))
        for i, t in enumerate(vc.argtypes):
            if t.startswith('0_'):
                if c_args[i] is None:
                    continue
                t = t.split('_')[1]
            if t == 'hint':
                if c_args[i] in ('context', 'flowable', 'quantity'):
                    continue
            elif t == 'float':
                if isinstance(c_args[i], float) or isinstance(c_args[i], int):
                    continue
            elif t == 'str':
                if isinstance(c_args[i], str):
                    continue
            elif t == 'direction':
                try:
                    check_direction(c_args[i])
                except KeyError:
                    raise ValueError('Argument %d [%s] is not a valid direction' % (i, c_args[i]))
                continue
            elif t == 'context':
                # check to ensure t is a recognized locally defined context
                cx = self._archive.tm[c_args[i]]
                if cx is None:
                    raise ValueError('Argument %d [%s] is not a recognized local context' % (i, c_args[i]))
                continue
            else:
                if not isinstance(c_args[i], str):
                    raise TypeError('%s [%d:%s]: Configuraton arguments must be strings and not entities' % (config, i,
                                                                                                             c_args[i]))
                e = self._archive.retrieve_or_fetch_entity(c_args[i])
                if e.entity_type == t:
                    continue
            raise TypeError('Argument %d should be type %s' % (i, t))
        return True


class ConfigureImplementation(CoreConfigureImplementation):

    _config_options = ('characterize_flow',)

    def characterize_flow(self, flow_ref, quantity_ref, value, location='GLO', overwrite=False, **kwargs):
        """
        A ConfigFlowCharacterization provides a procedural mechanism for specifying flow quantity characterizations
        after loading an archive.  The 'flow_ref' and 'quantity_ref' have to lookup successfully in the archive.

        Not clear how to deal with contexts

        :param flow_ref:
        :param quantity_ref:
        :param value:
        :param location:
        :param overwrite:
        :param kwargs:
        :return:
        """
        print('Characterizing flow %s by %s: %g' % (flow_ref, quantity_ref, value))
        flow = self._archive.retrieve_or_fetch_entity(flow_ref)
        qty = self._archive.retrieve_or_fetch_entity(quantity_ref)
        self._archive.tm.add_characterization(flow.link, flow.reference_entity, qty, value, context=flow.context,
                                              location=location,
                                              origin=self.origin, overwrite=overwrite)


class LcConfigureImplementation(ConfigureImplementation):

    _config_options = ('set_reference', 'unset_reference', 'characterize_flow', 'allocate_by_quantity')

    @staticmethod
    def _check_direction(pr, fl):
        exs = [x for x in pr.exchanges(flow=fl) if x.termination is None]
        if len(exs) == 2:
            raise ValueError('Direction must be specified')
        elif len(exs) == 0:
            raise KeyError('No unterminated exchanges found for flow %s' % fl)
        else:
            return exs[0].direction

    def set_reference(self, process_ref, flow_ref, direction=None, **kwargs):
        """
        A ConfigSetReference indicates that a particular exchange should be marked a reference exchange.  All terms are
        required.  The flow and direction must uniquely identify an exchange.

        :param process_ref:
        :param flow_ref:
        :param direction:
        :param kwargs:
        :return:
        """
        print('Setting reference %s [%s] for %s' % (flow_ref, direction, process_ref))
        pr = self._archive.retrieve_or_fetch_entity(process_ref)
        fl = self._archive.retrieve_or_fetch_entity(flow_ref)
        if direction is None:
            direction = self._check_direction(pr, fl)

        pr.set_reference(fl, direction)

    def unset_reference(self, process_ref, flow_ref, direction=None, **kwargs):
        """
        A ConfigBadReference provides a procedural mechanism for removing automatically-tagged reference flows, or for
        marking a byproduct as non-reference or non-allocatable.  The parts are 'process_ref', 'flow_ref', and
        'direction', but if process_ref is None, then all instances of the flow_ref and direction will be marked
        non-reference.

        Difficult to validate because any of the three arguments is allowed to be None (and even in nontrivial
        combinations)

        :param process_ref:
        :param flow_ref:
        :param direction:
        :param kwargs:
        :return:
        """
        print('UnSetting reference %s [%s] for %s' % (flow_ref, direction, process_ref))
        if process_ref is None:
            fl = self._archive.retrieve_or_fetch_entity(flow_ref)
            for p in self._archive.entities_by_type('process'):
                for x in p.references():
                    if x.flow is fl and x.direction == direction:
                        x.process.remove_reference(x.flow, x.direction)
        else:
            pr = self._archive.retrieve_or_fetch_entity(process_ref)
            if flow_ref is None:
                fds = [(x.flow, x.direction) for x in pr.references()]
                if direction is not None:
                    fds = [fd for fd in fds if fd[1] == direction]
                for fd in fds:
                    pr.remove_reference(*fd)
                return
            fl = self._archive.retrieve_or_fetch_entity(flow_ref)
            if direction is None:
                direction = self._check_direction(pr, fl)
            pr.remove_reference(fl, direction)

    def allocate_by_quantity(self, process_ref, quantity_ref, overwrite=False, **kwargs):
        """
        A ConfigAllocation provides a procedural mechanism for specifying quantity-wise allocations of processes at
        load time.  All that is required is a quantity; the process knows how to perform the allocation.  Note that
        reference flows not characterized with respect to the quantity will receive zero allocation.  So apply flow
        config first.

        :param process_ref:
        :param quantity_ref:
        :param overwrite:
        :param kwargs:
        :return:
        """
        print('Allocating %s by %s' % (process_ref, quantity_ref))
        p = self._archive.retrieve_or_fetch_entity(process_ref)
        qty = self._archive.retrieve_or_fetch_entity(quantity_ref)
        is_alloc = False
        if overwrite:
            for rf in p.reference_entity:
                p.remove_allocation(rf)
        else:
            for rf in p.reference_entity:
                is_alloc |= p.is_allocated(rf)
                '''
                try:
                    is_alloc |= p.is_allocated(rf)
                except MissingAllocation:
                    is_alloc = True
                    break
                '''

        # now apply the allocation
        if is_alloc:
            print('Allocation already detected for %s. Skipping this configuration.' % p)
        else:
            p.allocate_by_quantity(qty)



import os

from .data_source import DataSource, ResourceInfo, DataSourceError


TRACI_VERSIONS = {
    # Each entry must have ds_type and either a filename that exists or a download_url (optional download_md5sum)
    '2.1' : {
        'ds_type': 'Traci21Factors',
        'filename': 'traci_2_1_2014_dec_10_0.xlsx',
        'download_url': 'https://www.dropbox.com/s/p8w72o8iik5gzb6/traci_2_1_2014_dec_10_0.xlsx?dl=1',
        'download_md5sum': '4a16a40c453ab9084b610da7ded5e9c8'
    }  # would be great for this to be downloadable somewhere
}


E_CFG = {'hints': [  # cover elementary contexts that need directional hints
    ['context', 'air', 'to air'],
    ['context', 'water', 'to water']
]}


class TraciConfig(DataSource):

    _prefix = 'local.lcia.traci'
    _ifaces = ('index', 'quantity')

    @property
    def references(self):
        for k in TRACI_VERSIONS.keys():
            yield '.'.join([self._prefix, k])

    def interfaces(self, ref):
        if ref in self.references:
            for k in self._ifaces:
                yield k

    def _trim_origin(self, origin):
        gen = (_part for _part in origin.split('.'))
        for z in self._prefix.split('.'):
            if z != next(gen):
                raise ValueError('Invalid ref %s' % origin)
        return '.'.join(gen)

    def _check_info(self, info):
        if 'filename' in info:
            path = os.path.join(self._root, info['filename'])
            if os.path.exists(path):
                return ResourceInfo(path, info['ds_type'], None, None, E_CFG, None)
            else:
                print('Missing local source file: %s' % path)
            if 'download_url' not in info:
                raise DataSourceError('Missing source and no download link')

        return ResourceInfo(None, info['ds_type'],
                            info['download_url'], info.get('download_md5sum'), E_CFG, None)

    def make_resources(self, ref):
        """

        :param ref: can be a full reference OR a version (e.g. '2.1')
        :return:
        """
        if ref in TRACI_VERSIONS:
            ver = ref
            ref = '.'.join([self._prefix, ver])
        else:
            ver = self._trim_origin(ref)
            if ver not in TRACI_VERSIONS:
                raise KeyError('Unknown version %s' % ver)

        info = self._check_info(TRACI_VERSIONS[ver])
        yield self._make_resource(ref, info=info, interfaces=self._ifaces)



import re

from ...implementations import QuantityImplementation
from .q_info import q_info_21 as q_info


CAS_regexp = re.compile('^[0-9]{,6}-[0-9]{2}-[0-9]$')


def transform_numeric_cas(numeric_cas):
    if numeric_cas is None:
        return None
    if isinstance(numeric_cas, str):
        if numeric_cas == 'x':
            return None
        if bool(CAS_regexp.match(numeric_cas)):
            return numeric_cas
    ss = str(int(numeric_cas))
    return '-'.join([ss[:-3], ss[-3:-1], ss[-1:]])


class Traci21QuantityImplementation(QuantityImplementation):

    def get_canonical(self, name, **kwargs):
        if hasattr(name, 'entity_type'):
            if name.entity_type == 'quantity':
                qi = self[name]
                if qi is not None:
                    return qi
        for qi in self._archive.lcia_method_iter():
            if qi['Category'] == name:
                return qi
        return super(Traci21QuantityImplementation, self).get_canonical(name, **kwargs)

    def compartments(self, quantity=None, flowable=None, **kwargs):
        if quantity is not None:
            quantity = self.get_canonical(quantity)
        if flowable is not None:
            flowable = next(self._archive.row_for_key(flowable))
        comps = set()
        for col, val in q_info.items():
            if quantity is None or val.Category == quantity['Category']:
                if flowable is None:
                    comps.add(val.Compartment)
                else:
                    try:
                        cf = float(flowable[col])
                    except ValueError:
                        continue
                    if cf != 0:
                        comps.add(val.Compartment)
        for c in comps:
            yield c

    def factors(self, quantity, flowable=None, context=None, **kwargs):
        q = self.get_canonical(quantity)
        self._archive.add_method_and_compartment(method=q, compartment=context)
        return super(Traci21QuantityImplementation, self).factors(quantity, flowable=flowable, context=context)



from ..term_manager import TermManager, QuantityConflict
from ...entities import LcQuantity, LcFlow
from ...contexts import Context
import unittest


rq = LcQuantity.new('mass', 'kg', origin='test')
qq = LcQuantity.new('volume', 'm3', origin='test')


class TermManagerTest(unittest.TestCase):
    """
    This needs some serious development.

    Things to test:
    __getitem__: finds context by synonym; else returns None
    add_flow: multiple test cases:
     - add new item
     - add existing item (idempotency)
     - check reverse mapping
     - add conflicting item; prune
     - add conflicting item; merge
    add_cf-- why does this exist? We do need add_cfs for LciaEngine but not here
    add_characterization -- really need to specify use cases; use Traci experience to write tests and Ecoinvent LCIA to
     evaluate their robustness

    """
    @classmethod
    def setUpClass(cls):
        cls.tm = TermManager(quiet=False)
        cls.tm.add_context(['emissions', 'emissions to air', 'emissions to urban air'])

    def test_idempotent_context(self):
        """
        getitem(context) should just return the context
        :return:
        """
        cx = self.tm['emissions to air']
        self.assertIsInstance(cx, Context)
        self.assertIs(cx, self.tm[cx])

    def test_none_item(self):
        self.assertIs(self.tm[None], None)

    def test_undefined_item(self):
        for k in ('unspecified', 'UNKNOWN', 'Undefined', 'none'):
            self.assertIs(self.tm[k], self.tm._cm._null_entry)

    def test_add_flow(self):
        """
        We want to test several things:
         * extrac
        :return:
        """
        flow = LcFlow.new('Dummy Flow', ref_qty=rq, origin='test.origin')
        for k in ('phosphene', 'phxphn', '1234567'):
            flow._flowable.add_term(k)
        self.tm.add_flow(flow)
        self.assertEqual(self.tm._fm[flow.link], 'Dummy Flow')
        self.assertEqual(self.tm._fm['Dummy Flow'], 'Dummy Flow')
        with self.assertRaises(KeyError):
            self.tm._fm['1234567']

    def test_add_conflicting_contexts(self):
        c1 = self.tm.add_context(('emissions', 'emissions to water'))
        c2 = self.tm.add_context(('elementary flows', 'emissions to water'))
        self.assertIsNot(c1, c2)

    def test_add_flow_prune(self):
        pass

    def test_add_flow_merge(self):
        pass

    def test_add_flow_distinct(self):
        pass

    def test_add_characterization(self):
        cf = self.tm.add_characterization('water', rq, qq, 0.001)
        self.assertEqual(cf.value, .001)
        self.assertIs(self.tm.get_canonical('volume'), qq)

    ''' # can't figure out how to meaningfully trigger this error
    def test_dup_mass(self):
        self.tm.add_quantity(rq)
        dup_mass = LcQuantity('mass', referenceUnit='kg', Name='Mass', origin='test')
        with self.assertRaises(QuantityConflict):
            self.tm.add_quantity(dup_mass)
    '''


if __name__ == '__main__':
    unittest.main()



from __future__ import print_function, unicode_literals
import uuid

from antelope import Flow, CatalogRef, CONTEXT_STATUS_
from .entities import LcEntity
# from lcatools.entities.quantities import LcQuantity


class RefQuantityError(Exception):
    pass


def new_flow(name, ref_quantity, cas_number='', comment='', context=None, compartment=None, external_ref=None, **kwargs):
    if CONTEXT_STATUS_ == 'compat' and compartment is None:
        if context is None:
            compartment = []
        else:
            compartment = context.as_list()

    kwargs['CasNumber'] = kwargs.pop('CasNumber', cas_number)
    kwargs['Comment'] = kwargs.pop('Comment', comment)
    kwargs['Compartment'] = kwargs.pop('Compartment', compartment)

    if external_ref is None:
        return LcFlow.new(name, ref_quantity, **kwargs)
    return LcFlow(external_ref, Name=name, ReferenceQuantity=ref_quantity, context=context, **kwargs)


class LcFlow(LcEntity, Flow):

    _ref_field = 'referenceQuantity'
    _new_fields = ['CasNumber']  # finally abolishing the obligation for the flow to have a Compartment

    @classmethod
    def new(cls, name, ref_qty, **kwargs):
        """
        :param name: the name of the flow
        :param ref_qty: the reference quantity
        :return:
        """
        u = uuid.uuid4()
        return cls(str(u), Name=name, entity_uuid=u, ReferenceQuantity=ref_qty, **kwargs)

    def __setitem__(self, key, value):
        self._catch_flowable(key.lower(), value) or self._catch_context(key, value)
        super(LcFlow, self).__setitem__(key, value)

    @LcEntity.origin.setter
    def origin(self, value):  # pycharm lint is documented bug: https://youtrack.jetbrains.com/issue/PY-12803
        LcEntity.origin.fset(self, value)
        self._flowable.add_term(self.link)

    def __init__(self, external_ref, **kwargs):
        super(LcFlow, self).__init__('flow', external_ref, **kwargs)

        for k in self._new_fields:
            if k not in self._d:
                self._d[k] = ''
        self._chars_seen = dict()

    def make_ref(self, query):
        if self._query_ref is None:
            d = dict()
            for k in self.signature_fields():
                if k == self._ref_field:
                    continue
                if k in self._d:
                    d[k] = self._d[k]
            self._query_ref = CatalogRef.from_query(self.external_ref, query, self.entity_type,
                                                    uuid=self.uuid, **d)
            self._query_ref.context = self.context
            for k, v in self._chars_seen.items():
                self._query_ref.see_char(*k, v)

        return self._query_ref


    def __str__(self):
        cas = self.get('CasNumber')
        if cas is None:
            cas = ''
        if len(cas) > 0:
            cas = ' (CAS ' + cas + ')'
        context = '[%s]' % ';'.join(self.context)
        return '%s%s %s' % (self.get('Name'), cas, context)

    def characterize(self, quantity, value, context=None, origin=None, location='GLO', **kwargs):
        if context is None:
            context = self.context
        flowable = self.name
        if origin is None:
            origin = self.origin
        self.pop_char(quantity, context, location)
        return quantity.characterize(flowable, self.reference_entity, value, context=context, origin=origin,
                                     location=location, **kwargs)

    def get_context(self):
        return self.context

    def cf(self, quantity, **kwargs):
        if quantity.entity_type == 'quantity':
            return quantity.cf(self, **kwargs)
        elif quantity.entity_type == 'flow':
            return quantity.reference_entity.cf(self, **kwargs)
        else:
            raise TypeError('Invalid argument %s' % quantity)

    def see_char(self, qq, cx, loc, qrr):
        self._chars_seen[qq, cx, loc] = qrr
        if self._query_ref is not None:
            self._query_ref.see_char(qq, cx, loc, qrr)

    def chk_char(self, qq, cx, loc):
        return self._chars_seen[qq, cx, loc]

    def pop_char(self, qq, cx, loc):
        return self._chars_seen.pop((qq, cx, loc), None)



"""
autorange.py

Scales numerical values by powers of 1000 to present engineering notation. Also scales metric prefixes to match.
"""
from math import log10


metric_offsets = {
    'a': -6,
    'f': -5,
    'p': -4,
    'n': -3,
    'u': -2,
    'm': -1,
    'k': 1,
    'M': 2,
    'G': 3,
    'T': 4,
    'P': 5,
    'E': 6,
    'Z': 7
}


metric_prefixes = dict([(v, k) for k, v in metric_offsets.items()])
metric_prefixes[0] = ''


disallow_prefix = {'mol', 'MT'}  # units that start with these phrases should NOT use inferred prefix


class AutoRange(object):
    def _set_shift(self):
        self._shift = 0
        r = log10(self._range)
        while True:
            if r < 0:
                r += 3
                self._shift += 1
            elif r > 3:
                r -= 3
                self._shift -= 1
            else:
                return

    def __init__(self, rng):
        """
        creates an object for computing auto-ranged values.  The input argument should be the largest value that
        is expected to appear in the context.  It will be ranged to fall between 0-1000 (absolute value).
        :param rng:
        """
        self._range = abs(rng)
        if self._range == 0:
            print('Warning: autorange initialized with 0 range; using 1')
            self._range = 1
        self._shift = 0
        self._set_shift()

    @property
    def scale(self):
        return 10 ** (3 * self._shift)

    def adjust(self, value):
        return value * self.scale

    def adj_unit(self, unit):
        """
        This is the tricky, ad-hoc part.  Assume the first character in the unit string is a prefix IF it is found
        in the metric prefixes dictionary.  If it is not, assume it is unprefixed.
        :param unit: a unit string
        :return:
        """
        prefix = unit[0]

        if any([unit.startswith(x) for x in disallow_prefix]) or prefix not in metric_offsets:
            pre_shift = 0
            adj = unit
        else:
            pre_shift = metric_offsets[prefix]
            adj = unit[1:]

        post_shift = pre_shift - self._shift

        return metric_prefixes[post_shift] + adj



import unittest
import os

from ..ilcd import IlcdArchive

TEST_ARCHIVE = os.path.join(os.path.dirname(__file__), 'data', 'ilcd_test')
test_flow = 'f579de8c-8897-4bdb-9a0a-b36f8b13282e'

class IlcdTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.A = IlcdArchive(TEST_ARCHIVE, ref='test.ilcd')

    def test_archive_exists(self):
        self.assertEqual(self.A.ref, 'test.ilcd')

    def test_retrieve_flow(self):
        f = self.A.retrieve_or_fetch_entity(test_flow)
        self.assertEqual(f.entity_type, 'flow')
        self.assertEqual(f['Name'], 'RNA: natural gas, at consumer')

    def test_flowproperties(self):
        f = self.A.retrieve_or_fetch_entity(test_flow)
        cfs = [round(cf.value, 4) for cf in self.A.tm.factors_for_flowable(f.link)]
        self.assertEqual(len(cfs), 3)
        self.assertSetEqual(set(cfs), {0.0204, 0.015})

if __name__ == '__main__':
    unittest.main()



import unittest

from antelope import UnknownOrigin
from .test_aa_local import cat
from ..local import make_config, check_enabled

try:
    import antelope_background
    lci = True
except ImportError:
    lci = False

etypes = ('quantity', 'flow', 'process')

_debug = True

if __name__ == '__main__':
    _run_uslci = check_enabled('uslci')
else:
    _run_uslci = check_enabled('uslci') or _debug


if _run_uslci:
    cfg = make_config('uslci')
    try:
        gwp = next(cat.query('lcia.ipcc').lcia_methods())
    except UnknownOrigin:
        gwp = False


class UsLciTestContainer(object):

    class UsLciTestBase(unittest.TestCase):
        """
        Base class Mixin contains tests common to both ecospold and olca implemenations.  We hide it inside a container
        so that it does not get run automatically by unittest.main().  Then we can add implementation-specific tests in
        the subclasses.

        Thanks SO! https://stackoverflow.com/questions/1323455

        """
        _atype = None
        _initial_count = (0, 0, 0)
        _bg_len = None
        _ex_len = None
        _test_case_lcia = 0.0

        _petro_name = 'Petroleum refining, at refinery [RNA]'

        _petro_rx_values = set()


        @property
        def reference(self):
            return '.'.join([cfg.prefix, self._atype])

        @property
        def inx_reference(self):
            return '.'.join([self.reference, 'index'])

        @property
        def query(self):
            return cat.query(self.reference)

        def test_00_resources_exist(self):
            self.assertIn(self.reference, cat.references)

        def test_01_initial_count(self):
            ar = cat.get_archive(self.reference, strict=True)
            for i, k in enumerate(etypes):
                self.assertEqual(ar.count_by_type(k), self._initial_count[i])

        def test_10_index(self):
            inx_ref = cat.index_ref(self.reference, force=True)
            self.assertTrue(inx_ref.startswith(self.inx_reference))
            self.assertIn(inx_ref, cat.references)

        def _get_petro(self):
            return next(self.query.processes(Name='petroleum refining, at refinery'))

        def _preferred(self):
            yield self._get_petro()

        def test_12_get_petro(self):
            p = self._get_petro()
            self.assertEqual(p.name, self._petro_name)

        def test_20_inventory(self):
            p = self._get_petro()
            rx = [x for x in p.references()]
            inv = [x for x in p.inventory()]
            self.assertEqual(len(rx), len(self._petro_rx_values))
            self.assertEqual(len(inv), 51)

        def _get_fg_test_case_rx(self):
            p = next(self.query.processes(Name='Seedlings, at greenhouse, US PNW'))
            return p.reference()

        def _get_fg_test_case_lci(self):
            rx = self._get_fg_test_case_rx()
            return [x for x in self.query.lci(rx.process.external_ref, rx.flow.external_ref)]

        def test_21_exchange_relation(self):
            rx = self._get_fg_test_case_rx()
            k = next(self.query.flows(Name='CUTOFF Potassium fertilizer, production mix, at plant'))
            v = self.query.exchange_relation(rx.process.external_ref, rx.flow.external_ref, k.external_ref, 'Input')
            self.assertEqual(v, 0.000175)

        def test_22_petro_allocation(self):
            p = self._get_petro()
            self.assertEqual(len(p.reference_entity), len(self._petro_rx_values))
            rx_vals = set(round(next(p.exchange_values(rx.flow)).value, 6) for rx in p.references())
            self.assertSetEqual(rx_vals, self._petro_rx_values)

        @unittest.skipIf(lci is False, "no background")
        def test_30_bg_gen(self):
            preferred = list((rx.flow.external_ref, p.external_ref) for p in self._preferred() for rx in p.references())
            self.assertTrue(self.query.check_bg(reset=True, prefer=preferred))

        @unittest.skipIf(lci is False, "no background")
        def test_31_bg_length(self):
            self.assertEqual(len([k for k in self.query.background_flows()]), self._bg_len)
            self.assertEqual(len([k for k in self.query.exterior_flows()]), self._ex_len)

        @unittest.skipIf(lci is False, "no background")
        def test_32_lci_fg(self):
            lci = self._get_fg_test_case_lci()
            self.assertEqual(len(lci), 298 - self._bg_len)  # this works because the bg discrepancy shows up as cutoffs
            lead_vals = {1.5e-09, 2.3e-09, 0.0}
            self.assertSetEqual({round(x.value, 10) for x in lci if x.flow.name.startswith('Lead')}, lead_vals)

        @unittest.skipIf(lci is False, "no background")
        def test_40_lcia_fg(self):
            if gwp:
                lci = self._get_fg_test_case_lci()
                res = gwp.do_lcia(lci)
                self.assertAlmostEqual(res.total(), self._test_case_lcia)


class UsLciEcospoldTest(UsLciTestContainer.UsLciTestBase):

    _atype = 'ecospold'
    _initial_count = (5, 97, 5)
    _bg_len = 38
    _ex_len = 3285
    _test_case_lcia = 0.0415466  # more robust bc of ocean freight??

    _petro_rx_values = {0.037175, 0.049083, 0.051454, 0.051826, 0.059594, 0.061169, 0.112458, 0.252345, 0.570087}

    def test_get_by_id(self):
        f = self.query.get(2176)  # this flow was loaded via the config mechanism
        pvs = [k.value for k in f.profile()]
        self.assertGreaterEqual(len(pvs), 1)
        self.assertIn(11.111, pvs)


class UsLciOlcaTest(UsLciTestContainer.UsLciTestBase):

    _atype = 'olca'
    _initial_count = (8, 71, 3)  # 4 physical quantities + 4 alloc quantities
    _bg_len = 36
    _ex_len = 3680  # a "Diesel, at refinery" flow was incorrectly duplicated to exterior
    _test_case_lcia = .04110577

    # volume unit is m3 in olca, versus l in ecospold
    _petro_rx_values = {4.9e-05, 5.2e-05, 0.000112, 0.000252, 0.00057, 0.037175, 0.051454, 0.059594, 0.061169}

    def _preferred(self):
        yield self._get_petro()
        yield self.query.get('cdc143eb-fff8-3618-85cd-bce83d96390f')  # veneer, at veneer mill, preferred for wood fuel


if __name__ == '__main__':
    unittest.main()



"""
With inspiration from bw2data/io/import_ecospold2.py and affiliated files

The purpose of this archive (and ILCD alike) is to provide a common interface, a la the semantic
web paper, to a collection of process data, to wit:
 - list of process metadata:
   UUID | Name | Spatial | Temporal | Ref Product | Comment |

"""

from __future__ import print_function, unicode_literals

import uuid
import re
import os

from lxml import objectify
# from lxml.etree import tostring

from ..archives import LcArchive
from ..entities import LcQuantity, LcFlow, LcProcess
# from lcatools.interact import parse_math
# from lcatools.exchanges import DirectionlessExchangeError

from .file_store import FileStore
from .xml_widgets import find_tag
from .parse_math import parse_math

tail = re.compile('/([^/]+)$')

"""
Used to install conversion factors between different flow reference units.  Satisfies the requirement:
conversion_dict[(k1, k2)] = f implies 1 k1 = f k2

Note: one conversion factor that was omitted is required by the 2015-era US LCI database: The process "Biodegradable
loose fill [RNA]" requires "natural gas, combusted in equipment" measured in kWh, but there is no conversion from m3 of
natural gas combusted to kWh.  My best guess is on a GCV basis of fuel input, so ~40 MJ/m3 = 11.111 kWh / m3

This value must be entered by hand when the USLCI database is loaded.
"""
conversion_dict = {
    ('kBq', 'Bq'): 1000,
    ('t', 'kg'): 1000,
    ('kg*km', 't*km'): .001,
    ('ha', 'm2'): 10000,
    ('sh tn', 't'): 0.907185,
    ('sh tn', 'kg'): 907.185,
    ('kWh', 'MJ'): 3.6,
    ('MJ', 'btu'): 947.817,
    ('m3', 'l'): 1000
}

def apply_conversion(local_q, f):
    if (f.unit, local_q.unit) in conversion_dict:
        val = conversion_dict[(f.unit, local_q.unit)]
        f.reference_entity['UnitConversion'][local_q.unit] = val
        local_q['UnitConversion'][f.unit] = 1.0 / val
        return local_q.cf(f) == val
    elif (local_q.unit, f.unit) in conversion_dict:
        val = conversion_dict[(local_q.unit, f.unit)]
        local_q['UnitConversion'][f.unit] = val
        f.reference_entity['UnitConversion'][local_q.unit] = 1.0 / val
        return local_q.cf(f) == 1.0 / val
    return False


def not_none(x):
    return x if x is not None else ''


class EcospoldVersionError(Exception):
    pass


class EcospoldV1Archive(LcArchive):
    """
    Create an Ecospold Archive object from a path.  By default, assumes the path points to a literal
    .7z file, of the type that one can download from the ecoinvent website.  Creates an accessor for
    files in that archive and allows a user to
    """

    nsmap = 'http://www.EcoInvent.org/EcoSpold01'  # only valid for v1 ecospold files
    spold_version = tail.search(nsmap).groups()[0]

    def __init__(self, source, prefix=None, ns_uuid=None, **kwargs):
        """
        Just instantiates the parent class.
        :param source: physical data source
        :param prefix: difference between the internal path (ref) and the ILCD base
        :param ns_uuid: NS UUID not allowed for ecospold ve
        :return:
        """
        if ns_uuid is None:
            ns_uuid = uuid.uuid4()
        super(EcospoldV1Archive, self).__init__(source, ns_uuid=ns_uuid, **kwargs)
        if prefix is not None:
            self._serialize_dict['prefix'] = prefix
        self._q_dict = dict()
        self._archive = FileStore(self.source, internal_prefix=prefix)

    def list_datasets(self):
        assert self._archive.remote is False, "Cannot list objects for remote archives"
        for f in self._archive.listfiles():
            yield f

    def _fetch_filename(self, filename):
        """
        In USLCI (honestly, I've never seen another ecospoldv1 archive so I can't generalize), activities  whose
        names contain '<' or '>' characters have those characters stripped from the filenames, so we try that here as
        a workaround.
        :param filename:
        :return:
        """
        try:
            s = self._archive.readfile(filename)
        except KeyError:
            trim_lg = ''.join(k for k in filename if k not in '<>')
            s = self._archive.readfile(trim_lg)
        return s

    def _get_objectified_entity(self, filename):
        o = objectify.fromstring(self._fetch_filename(filename))
        if o.nsmap[None] != self.nsmap:
            raise EcospoldVersionError('This class is for EcoSpold v%s only!' % self.nsmap[-2:])
        return o

    def _create_quantity(self, unitstring):
        """
        In ecospold v1, quantities are only units, defined by string
        :param unitstring:
        :return:
        """
        if unitstring in self._q_dict:
            q = self._q_dict[unitstring]
        else:
            ref_unit, _ = self._create_unit(unitstring)

            q = LcQuantity(unitstring, Name='EcoSpold Quantity %s' % unitstring,
                           ReferenceUnit=ref_unit, Comment=self.spold_version)
            self.add(q)

            self._q_dict[unitstring] = q

        return q

    def _create_flow(self, exch):
        """
        An ecospold01 exchange is really just a long attribute list, plus an inputGroup or outputGroup (ignored here)
        :param exch:
        :return:
        """
        number = str(exch.get('number'))
        try_f = self[number]
        if try_f is not None:
            f = try_f
            assert f.entity_type == 'flow', "Expected flow, found %s" % f.entity_type

        else:
            # generate flow
            n = exch.get("name")
            q = self._create_quantity(exch.get("unit"))
            c = not_none(exch.get("generalComment"))
            cas = not_none(exch.get("CASNumber"))
            cat = [k for k in filter(None, (exch.get('category'), exch.get('subCategory')))]

            f = LcFlow(number, Name=n, CasNumber=cas, Comment=c, Compartment=cat, ReferenceQuantity=q)
            self.add(f)

        if exch.get("unit") != f.unit:
            local_q = self._create_quantity(exch.get("unit"))
            if local_q.cf(f) == 0.0:
                if not apply_conversion(local_q, f):
                    print('Flow %s needs characterization for unit %s' % (f, local_q))
                    val = parse_math(input('Enter conversion factor 1 %s = x %s' % (f.unit, local_q)))
                    self.tm.add_characterization(f.link, f.reference_entity, local_q, val, context=f.context,
                                                 origin=self.ref)
        return f

    def _extract_exchanges(self, o):
        rf = set()  # reference flows
        flowlist = []

        for exch in o.dataset.flowData.getchildren():
            f = self._create_flow(exch)
            if hasattr(exch, 'outputGroup'):
                d = 'Output'
                if exch.outputGroup == 0 or exch.outputGroup == 2:
                    rf.add(f)
            elif hasattr(exch, 'OutputGroup'):
                d = 'Output'
                if exch.OutputGroup == 0 or exch.OutputGroup == 2:
                    rf.add(f)
            elif hasattr(exch, 'inputGroup'):
                d = 'Input'
            elif hasattr(exch, 'InputGroup'):
                d = 'Input'
            else:
                print('Abandoning directionless exchange for flow %s' % f)
                continue
                # raise DirectionlessExchangeError(tostring(exch))
            local_q = self._create_quantity(exch.get("unit"))
            v = float(exch.get('meanValue'))  # returns none if missing
            if local_q is not f.reference_entity:
                v = v / local_q.cf(f)
            c = exch.get('generalComment')
            flowlist.append((f, d, v, c))
        return rf, flowlist

    def _try_exts(self, basename):
        exts = ('', '.xml', '.XML', '.spold', '.SPOLD')
        for e in exts:
            try:
                return self._get_objectified_entity(basename + e)
            except FileNotFoundError:
                pass
        raise KeyError(basename)

    def _create_process(self, filename):
        """
        Extract dataset object from XML file
        :param filename: filename with or without extension.  various extensions .xml .XML .spold .SPOLD are tried
        if the supplied name fails.
        external_ref should not have file extension.
        :return:
        """
        '''
        first sanitize the input
        '''
        if filename.lower().endswith('.xml') or filename.lower().endswith('.spold'):
            ext_ref, ext = os.path.splitext(filename)
        else:
            ext_ref = filename
        try_p = self[ext_ref]
        if try_p is not None:
            p = try_p
            assert p.entity_type == 'process', "Expected process, found %s" % p.entity_type
            return p

        o = self._try_exts(ext_ref)

        p_meta = o.dataset.metaInformation.processInformation
        n = p_meta.referenceFunction.get('name')

        # create new process
        g = p_meta.geography.get('location')
        stt = {'begin': str(find_tag(p_meta, 'startDate')), 'end': str(find_tag(p_meta, 'endDate'))}

        c = p_meta.referenceFunction.get('generalComment')

        cls = [p_meta.referenceFunction.get('category'), p_meta.referenceFunction.get('subCategory')]
        p = LcProcess(ext_ref, Name=n, Comment=c, SpatialScope=g, TemporalScope=stt,
                      Classifications=cls)

        rf, flowlist = self._extract_exchanges(o)

        for flow, f_dir, val, cmt in flowlist:
            if flow in rf and f_dir == 'Output':
                term = None
            else:
                term = self.tm[flow.context]
            self._print('Exch %s [%s] (%g)' % (flow, f_dir, val))
            x = p.add_exchange(flow, f_dir, reference=None, value=val, termination=term, add_dups=True)
            if cmt is not None:
                x.comment = cmt

        for ref in rf:
            p.set_reference(ref, 'Output')

        self.add(p)

        return p

    def _fetch(self, key, **kwargs):
        """
        If the argument is an external reference for a process, try loading the file
        :param key:
        :param kwargs:
        :return:
        """
        try:
            self._create_process(key)
            return self[key]
        except KeyError:
            raise KeyError('No way to fetch key "%s". try load_all()' % key)

    def _load_all(self):
        """
        USLCI Hack: the flow "natural gas, combusted in equipment" is used in the database with three different
        reference units (kWh, l, and m3).  We want the reference unit to be m3.  So.. we load the process with
        that characterization first.
        :return:
        """
        for x in self.list_datasets():
            if bool(re.search('Natural gas, combusted in industrial equipment', x)):
                self.retrieve_or_fetch_entity('Natural gas, combusted in industrial equipment')
        for k in self.list_datasets():
            self._create_process(k)
        self.check_counter('quantity')
        self.check_counter('flow')
        self.check_counter('process')



from antelope_core.data_sources.data_source import DataSource, ResourceInfo

VALID_FORMATS = ('olca', 'ecospold')  # 'ecospold')


INFO = {
    'olca': ResourceInfo(None,
                        'OpenLcaJsonLdArchive',
                         'https://www.dropbox.com/s/upuo0yc7hwevmvs/NREL_USLCI_JSON__20180109.zip?dl=1',
                         '3e8471b1a73b38013c3ab8e8b2df0ef4',
{
    "unset_reference": [
          [
            "dc72e285-719b-318b-9c9c-c838846a9cf4",
            "d939590b-a0d7-310c-8952-9921ed64a078",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",  # activity-as-reference
            "471bcc19-36f7-3ad8-b3ae-e39d2ab5fe44",
            "Output"
          ],
          [
              '155839ec-d6ab-3ed6-b7c8-45586e676f6d',  # ambiguous termination
              '9dfc9e7c-3c20-3f4d-9d79-d54ec2d8d08b',
              'Output'
          ]
    ],
    "set_reference": [
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "d939590b-a0d7-310c-8952-9921ed64a078",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "c4069217-dfd4-324a-9243-2ee8058809d6",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "0e44e579-abb0-3c77-af64-c774d65be529",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "c5f94bb1-b39e-39fa-b616-376a30531c2d",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "4f1bfabb-f0d3-3f1f-a905-2fd9b36bbfa0",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "562918fe-3ff5-33f4-abaa-d8c615380d25",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "5f96ad43-abab-3b62-bcf8-c71f9d6c7881",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "30a64e63-3583-3f83-8d8d-0b6ea003b888",
            "Output"
          ],
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "588384cf-851f-391e-bf49-c6e5309d76db",
            "Output"
          ]
    ],
    "characterize_flow": [
          [
            "0e44e579-abb0-3c77-af64-c774d65be529",
            "93a60a56-a3c8-11da-a746-0800200b9a66",
            750
          ],
          [
            "562918fe-3ff5-33f4-abaa-d8c615380d25",
            "93a60a56-a3c8-11da-a746-0800200b9a66",
            975
          ],
          [
            "588384cf-851f-391e-bf49-c6e5309d76db",
            "93a60a56-a3c8-11da-a746-0800200b9a66",
            500
          ],
          [
            "c4069217-dfd4-324a-9243-2ee8058809d6",
            "93a60a56-a3c8-11da-a746-0800200b9a66",
            0.737
          ],
          [
            "c5f94bb1-b39e-39fa-b616-376a30531c2d",
            "93a60a56-a3c8-11da-a746-0800200b9a66",
            809
          ],
          [
            "d939590b-a0d7-310c-8952-9921ed64a078",
            "93a60a56-a3c8-11da-a746-0800200b9a66",
            850
          ]
    ],
    "allocate_by_quantity": [
          [
            "0aaf1e13-5d80-37f9-b7bb-81a6b8965c71",
            "93a60a56-a3c8-11da-a746-0800200b9a66"
          ]
    ],
    "hints": [
     ["context", "water", "to water"],
     ["context", "air", "to air"],
     ["quantity", "Energy", "Net Calorific Value"]
    ]

                         },
                         {}),
    'ecospold': ResourceInfo(None,
                             'EcospoldV1Archive',
                             'https://www.dropbox.com/s/xcsxgj5pipgwe43/USLCI_Processes_ecospold1.zip?dl=1',
                             'f022051ee87c0976e126ebcc2bced017',
{
    "set_reference": [
        ['Natural gas, combusted in industrial equipment', 2176, 'Output']  # hack to force load of flow 2176
    ],
    "unset_reference": [
        ['Crude oil, in refinery', None, 'Output'],  # remove all Output references
#        [None, 9218, 'Output'],  # recovered energy, generic  # can't sort 'None' :(
        ['Petroleum refining, at refinery', 32123, 'Output'],  # documentary activity flow -- not allocatable
        ['Bucked and debarked log, hardwood, green, at veneer mill, E', 5339, 'Output'],  # ambiguous term error
        ['Harvesting, fresh fruit bunch, at farm (2)', 23293, 'Output']  # terrible naming practice
    ],
    "characterize_flow": [
        [775, 'kg', 0.867],  # 1.153 L/kg from ecospold doc
        [869, 'kg', 0.739],  # 1.353 L/kg from ecospold doc
        [16694, 'kg', 0.543],  # 1.84 L/kg from ecospold doc
        [16696, 'kg', 0.944],  # 1.059 L/kg from ecospold doc
        [16700, 'kg', 0.809],  # 1.236 L/kg from ecospold doc
        [16702, 'kg', 0.737],  # 1.356 m3/kg from ecospold doc
        [2176, 'kWh', 11.111]   # see ecospold documentation
    ],
    "allocate_by_quantity": [
        [
            'Petroleum refining, at refinery', 'kg'
        ]
    ],
    "hints": [
        [ "context", "water", "to water" ],
        [ "context", "air", "to air" ],
        [ "quantity", "EcoSpold Quantity m3", "volume" ],
        [ "quantity", "EcoSpold Quantity kg", "mass" ]
    ]
},
                             {
                                 'prefix': 'USLCI_Processes_ecospold1/USLCI_Processes_ecospold1',
                                 'ns_uuid': '96386cae-b651-47ca-8fcd-d3a1aebd6034'
                             })}


class UsLciConfig(DataSource):

    prefix = 'local.uslci'
    _ifaces = ('exchange', 'quantity')

    def _ref(self, fmt):
        return '.'.join([self.prefix, fmt])

    @property
    def references(self):
        for f in VALID_FORMATS:
            yield self._ref(f)

    def interfaces(self, ref):
        for k in self._ifaces:
            yield k

    def make_resources(self, ref):
        if ref not in self.references:
            raise ValueError('Unknown reference %s' % ref)
        fmt = ref.split('.')[-1]
        info = INFO[fmt]
        yield self._make_resource(ref, info=info, interfaces=self._ifaces)



from ..entities import LcEntity
import unittest
from uuid import uuid4


class LcEntityTest(unittest.TestCase):
    def test_init(self):
        self.assertIsInstance(LcEntity('goober', uuid4()), LcEntity)

    def test_case(self):
        e = LcEntity('my goober', uuid4(), Quorum=42)
        self.assertEqual(e['quorum'], 42)
        self.assertEqual(e['comment'], '')

    def test_null_property(self):
        e = LcEntity('my sharona', uuid4(), Shame=None)
        self.assertFalse(e.has_property('Shame'))

    def test_contains(self):
        e = LcEntity('the biggest goober', uuid4(), Annotation='a subtle yet profound statement', domain='society')
        self.assertTrue(e.has_property('annotation'))
        self.assertTrue(e.has_property('Annotation'))
        self.assertTrue(e.has_property('dOmAiN'))
        self.assertTrue(e.has_property('dOmAiN'))
        self.assertFalse(e.has_property('SoCiEtY'))
        self.assertFalse(e.has_property('annotat'))


if __name__ == '__main__':
    unittest.main()



"""
Import ecospold2 files
"""

from __future__ import print_function, unicode_literals

from collections import namedtuple
from time import time

from lxml import objectify
from lxml.etree import XMLSyntaxError

from ...characterizations import QRResult
from ...entities import LcQuantity, LcFlow, LcProcess
from ...entities.processes import AlreadyAReference
from ...exchanges import ExchangeValue, DirectionlessExchangeError
from ...lcia_results import LciaResult
from ..ecospold import tail
from ..file_store import FileStore
from ...archives import LcArchive, uuid_regex
from ..xml_widgets import *

from .ecospold2_index import EcoSpold2IndexImplementation
from .master_data import EcoSpoldMasterData


EcospoldExchange = namedtuple('EcospoldExchange', ('flow', 'direction', 'value', 'termination', 'is_ref', 'comment'))
EcospoldLciaResult = namedtuple('EcospoldLciaResult', ('Method', 'Category', 'Indicator', 'score'))


def spold_reference_flow(filename):
    """
    second UUID, first match should be reference flow uuid
    :param filename:
    :return: first uuid, second uuid
    """
    m = uuid_regex.findall(filename)
    try:
        return m[0][0], m[1][0]
    except IndexError:
        try:
            return m[0][0], None
        except IndexError:
            print('No UUID found in %s' % filename)
            raise


class EcospoldV2Error(Exception):
    pass


def _add_syn_if(syn, synset):
    g = syn.strip()
    if g != '' and g != 'PSM':
        synset.add(syn)


def _syn_set(exch):
    synonym_tag = list(find_tags(exch, 'synonym'))
    syns = set()
    if len(synonym_tag) == 1:
        # parse the comma-separated list
        if bool(re.search('etc\.', str(synonym_tag[0]))):
            syns.add(str(synonym_tag[0]).strip())
        else:
            for x in str(synonym_tag[0]).split(', '):
                _add_syn_if(x, syns)
    else:
        # multiple entries- allow embedded comma-space
        for syn in synonym_tag:
            _add_syn_if(str(syn), syns)
    return syns


class EcospoldV2Archive(LcArchive):
    """
    class for loading metadata from ecospold v2 files. Now I know ecoinvent supplies a whole ton of supplementary
    information in files that are *outside* the ecospold archives- and that information is going to be IGNORED.
    or loaded separately. But not handled here.
    """

    nsmap = 'http://www.EcoInvent.org/EcoSpold02'  # only valid for v1 ecospold files
    spold_version = tail.search(nsmap).groups()[0]

    def __init__(self, source, prefix=None, linked=True, **kwargs):
        """
        Just instantiates the parent class.
        :param source: physical data source
        :param prefix: relative path for datasets from the archive root
        :param linked: [True] whether the archive includes unlinked or linked datasets. Reference exchanges
        get detected differently in one case versus the other (see _create_process)
        :return:
        """
        super(EcospoldV2Archive, self).__init__(source, **kwargs)
        if prefix is not None:
            self._serialize_dict['prefix'] = prefix

        self._archive = FileStore(self.source, internal_prefix=prefix)
        self._master = EcoSpoldMasterData(self.source)

        ''' # not sure why I did this (preload all elem flows) but now I am taking it out
        for exch in self._master.elementary_exchanges.values():
            self._create_flow(exch)
        '''
        print('Warning: not preloading flows')

        self._linked = linked
        self._process_flow_map = defaultdict(set)
        self._terminations = defaultdict(set)
        self._map_datasets()

    @property
    def master_data(self):
        return self._master

    def load_flows(self):
        for exch in self._master.elementary_exchanges.values():
            self._create_flow(exch)

    '''
    def fg_proxy(self, proxy):
        for ds in self.list_datasets(proxy):
            self.retrieve_or_fetch_entity(ds)
        return self[proxy]

    def bg_proxy(self, proxy):
        return self.fg_proxy(proxy)
    '''
    def make_interface(self, iface, privacy=None):
        if iface == 'index':
            return EcoSpold2IndexImplementation(self)
        return super(EcospoldV2Archive, self).make_interface(iface)

    # no need for _ref_to_key - keys in ecospold are uuids
    def _fetch_filename(self, filename):
        st = self._archive.readfile(filename)
        if st is None:
            raise FileNotFoundError
        return st

    def _map_datasets(self):
        for f in self.list_datasets():
            p, r = spold_reference_flow(f)
            self._process_flow_map[p].add(r)
            self._terminations[r].add(p)

    @property
    def ti(self):
        return self._terminations

    def count_by_type(self, entity_type):
        if entity_type == 'process':
            return len(self._process_flow_map)
        return super(EcospoldV2Archive, self).count_by_type(entity_type)

    def processes(self, **kwargs):
        if len(kwargs) == 0:
            for p in self._process_flow_map.keys():
                yield self.retrieve_or_fetch_entity(p)
        else:
            for p in self.search('process', **kwargs):
                yield p

    def list_datasets(self, startswith=None):
        assert self._archive.remote is False, "Cannot list objects for remote archives"
        for x in self._archive.listfiles(in_prefix=startswith):
            yield x

    def _get_objectified_entity(self, filename):
        try:
            o = objectify.fromstring(self._fetch_filename(filename))
        except FileNotFoundError:
            raise
        except ValueError:
            print('failed on :%s:' % filename)
            return None
        if o.nsmap[None] != self.nsmap:
            raise EcospoldV2Error('This class is for EcoSpold v%s only!' % self.nsmap[-2:])
        return o

    def _get_objectified_entity_with_lt_gt(self, filename):
        try:
            f = self._fetch_filename(filename)
            f = re.sub(' < ', ' &lt; ', re.sub(' > ', ' &gt; ', f.decode()))
            o = objectify.fromstring(f)
        except ValueError:
            print('failed on :%s:' % filename)
            return None
        except TypeError:
            print('failed on :%s:' % filename)
            raise
        if o.nsmap[None] != self.nsmap:
            raise EcospoldV2Error('This class is for EcoSpold v%s only!' % self.nsmap[-2:])
        return o

    def _create_quantity(self, exchange):
        """
        In ecospold v2, quantities are defined by a Properties.xml master file-- but unfortunately Ecospold files
        do not USE them to describe exchanges-- only to characterize the flows.  Exchanges themselves are still defined
        using only a unit of measure, not a quantity.

        Moreover, there are widespread inconsistencies in the implementation of Ecoinvent's units and flow properties,
        which seem largely crafted for ad hoc convenience (e.g.: "net heating value" measured in MJ/kg and "heating
        value, net" measured in MJ).

        Thereby, quantities in Ecospold v2 are still only units, defined by string.  They do get their own uuids, but only
        as 'properties' of the flows- flows themselves are only measured by unit.
        this code is cc'd from ecospold1
        :param exchange:
        :return:
        """
        unitstring = exchange.unitName.text
        unit_uuid = exchange.attrib['unitId']
        try_q = self[unit_uuid]
        if try_q is None:
            ref_unit, _ = self._create_unit(unitstring)

            q = LcQuantity(unit_uuid, Name='EcoSpold Quantity %s' % unitstring, ReferenceUnit=ref_unit,
                           Comment=self.spold_version)
            self.add(q)
        else:
            q = try_q

        return q

    @staticmethod
    def _cls_to_text(i):
        if isinstance(i, objectify.ObjectifiedElement):
            return ': '.join([i.classificationSystem.text, i.classificationValue.text])
        else:
            return ''

    @staticmethod
    def _cat_to_text(i):
        if isinstance(i, objectify.ObjectifiedElement):
            return [i.compartment.text, i.subcompartment.text]
        else:
            return []

    def _create_flow(self, exchange):
        """
        makes a flow entity and adds to the db
        :param exchange:
        :return:
        """
        if 'intermediate' in exchange.tag:
            try:
                uid = exchange.attrib['intermediateExchangeId']
            except KeyError:
                uid = exchange.attrib['id']
            cat = [self._cls_to_text(exchange.classification)]
        elif 'elementary' in exchange.tag:
            try:
                uid = exchange.attrib['elementaryExchangeId']
            except KeyError:
                uid = exchange.attrib['id']
            cat = self._cat_to_text(exchange.compartment)
        else:
            raise AttributeError('No exchange type found for id %s' % exchange.attrib['id'])

        f = self[uid]
        if f is not None:
            return f

        if 'casNumber' in exchange.attrib:
            cas = exchange.attrib['casNumber']
        else:
            cas = ''

        q = self._create_quantity(exchange)

        n = exchange.name.text
        c = 'EcoSpold02 Flow'

        f = LcFlow(uid, Name=n, CasNumber=cas, Comment=c, Compartment=cat, ReferenceQuantity=q)
        # TODO: implement ecospold flow properties, only for reference products

        syns = _syn_set(exchange)
        syns.add(uid)
        f['Synonyms'] = syns

        self.add(f)

        return f

    @staticmethod
    def _get_process_comment(ad, ud):
        if 1:
            gc = find_tag(ad, 'generalComment')
            if gc == '':
                # print('activity ID %s: no comment' % ud)
                return 'no comment.'
            try:
                c = render_text_block(gc)
            except KeyError:
                print(ud)
                raise
            except XmlWidgetError:
                print(ud)
                raise
            return c
        try:
            c = find_tag(ad, 'generalComment')['text'].text
        except TypeError:
            c = 'no comment.'
        except AttributeError:
            print('activity ID %s: no comment' % ud)
            c = 'no comment.'
        return c

    def _create_process_entity(self, o):
        """
        Constructs the process without populating exchanges
        :param o:
        :return:
        """
        ad = find_tag(o, 'activityDescription')

        u = ad.activity.get('id')

        if self[u] is not None:
            return self[u]

        n = find_tag(ad, 'activityName').text

        c = self._get_process_comment(ad, u)

        g = find_tag(ad, 'geography').shortname.text

        tp = find_tag(ad, 'timePeriod')
        stt = {'begin': tp.get('startDate'), 'end': tp.get('endDate')}
        cls = [self._cls_to_text(i) for i in find_tags(ad, 'classification')]

        p = LcProcess(u, Name=n, Comment=c, SpatialScope=g, TemporalScope=stt,
                      Classifications=cls)

        parent = find_tag(o, 'activity').get('parentActivityId')
        if parent is not None:
            p['ParentActivityId'] = parent

        self.add(p)
        return p

    def _grab_reference_flow(self, o, rf_uuid):
        """
        Create a reference exchange from the flowdata
        :param o:
        :param rf_uuid:
        :return:
        """
        for x in find_tag(o, 'flowData').getchildren():
            if 'intermediate' in x.tag:
                if x.attrib['intermediateExchangeId'] == rf_uuid:
                    flow = self._create_flow(x)
                    value = float(x.get('amount'))  # or None if not found
                    return flow, value

        raise KeyError('Noted reference exchange %s not found!' % rf_uuid)

    def _collect_exchanges(self, o, ref_uuid):
        """

        :param o:
        :param ref_uuid: strictly for diagnostic purposes
        :return:
        """
        flowlist = []

        for exch in find_tag(o, 'flowData').getchildren():
            if 'parameter' in exch.tag:
                continue
            if 'impactIndicator' in exch.tag:
                continue

            f = self._create_flow(exch)
            is_ref = False
            if hasattr(exch, 'outputGroup'):
                d = 'Output'
                og = exch.outputGroup
                if og == 0:
                    is_ref = True
                elif og == 2:  # 1, 3 not used
                    for cls in find_tags(exch, 'classification'):
                        if cls.classificationSystem == 'By-product classification':
                            if str(cls.classificationValue).startswith('allocat'):
                                is_ref = True
            elif hasattr(exch, 'inputGroup'):
                d = 'Input'
            else:
                raise DirectionlessExchangeError
            v = float(exch.get('amount'))  # or None if not found
            t = exch.get('activityLinkId')  # or None if not found
            if t is None:
                if 'elementary' in exch.tag:
                    t = self.tm[f.context]
            try:
                c = '; '.join([str(c) for c in exch.iterchildren() if c.tag == '{%s}comment' % o.nsmap[None]])
            except ValueError:
                ad = find_tag(o, 'activityDescription')

                u = ad.activity.get('id')

                print('Failed reading comment %s_%s: %s' % (u, ref_uuid, exch.get('id')))
                c = None
            flowlist.append(EcospoldExchange(f, d, v, t, is_ref, c))
        return flowlist

    @staticmethod
    def _collect_impact_scores(o):  # , process, flow):
        """
        the old "1115"
        :param o:
        :return:
        """

        scores = []
        # exch = ExchangeValue(process, flow, 'Output', value=1.0)

        for cf in find_tag(o, 'flowData').getchildren():
            if 'impactIndicator' in cf.tag:
                m = cf.impactMethodName.text
                c = cf.impactCategoryName.text
                x = cf.name.text
                v = float(cf.get('amount'))
                scores.append(EcospoldLciaResult(m, c, x, v))

        return scores

    def objectify(self, process_uuid, rf_uuid):
        filename = '%s_%s.spold' % (process_uuid.lower(), rf_uuid.lower())
        self._print('\nObjectifying %s' % filename)
        try:
            o = self._get_objectified_entity(filename)
        except FileNotFoundError:
            raise
        except XMLSyntaxError:
            print('  !!XMLSyntaxError-- trying to escape < and > signs')
            try:
                o = self._get_objectified_entity_with_lt_gt(filename)
            except XMLSyntaxError:
                print('  !!Failed loading %s' % filename)
                raise
        return o

    def _create_process_and_single_reference(self, process_uuid, ref_uuid, exchanges=True):
        """
        Extract dataset object from XML file.  Handles unlinked and linked archives differently:
         * for unlinked (unallocated) archives, there is exactly one spold file for each process.  Reference flows are
           determined contextually, based on outputGroup and By-product classification (see _collect_exchanges).

         * for linked (allocated) archives, there is a distinct spold file for each allocation. References are
           determined based on the filename, which specifies the reference exchange for which the allocation was
           performed.

        :param process_uuid: uuid of activity
        :param ref_uuid: uuid of reference flow
        :return:
        """
        p = self[process_uuid]
        if p is not None:
            if p.has_reference(ref_uuid):
                return p

        try:
            o = self.objectify(process_uuid, ref_uuid)
        except KeyError:
            raise FileNotFoundError

        p = self._create_process_entity(o)

        if p.has_reference(ref_uuid):
            self._print('Process %s already has reference %s' % (process_uuid, ref_uuid))
            return p

        '''
        if self._linked:
            rf, value = self._grab_reference_flow(o, ref_uuid)
            rx = p.add_exchange(rf, 'Output', value=value)  # this should get overwritten with an ExchangeValue later
            p.set_reference(rf, 'Output')
            self._print('# Identified reference exchange\n %s' % rx)
        else:
            rx = None
        '''
        if exchanges:
            rx = None
            exchs = self._collect_exchanges(o, ref_uuid)
            if self._linked:
                for exch in exchs:
                    if exch.is_ref and exch.value != 0:
                        try:
                            _rx = p.add_exchange(exch.flow, exch.direction, value=exch.value)
                        except AlreadyAReference:
                            raise EcospoldV2Error('Already a reference: %s | %s' % (process_uuid, ref_uuid))
                        p.set_reference(exch.flow, exch.direction)
                        self._print('# Identified reference exchange\n %s' % _rx)
                        if len(exch.comment) > 0:
                            _rx.comment = exch.comment
                        if exch.flow.external_ref == ref_uuid:
                            if rx is None:
                                rx = _rx
                            else:
                                raise EcospoldV2Error('Multiple rx found: %s | %s' % (process_uuid, ref_uuid))
                if rx is None:
                    raise EcospoldV2Error('No rx found: %s | %s' % (process_uuid, ref_uuid))
            for exch in exchs:
                """
                If the dataset is linked, all we do is load non-zero exchanges, ideally all with terminations.  Spurious
                 terminations in reference exchanges are dropped (deprecated EI linker feature)
                If the dataset is unlinked, we will err on the side of adding zero-valued exchanges.  The unlinked data
                 should include as much information as possible.  Spurious terminations in reference exchanges are
                  retained, but the reference status is dropped (in-use EI linker feature)

                We could simplify this function by placing those tests in _collect_exchanges, but I would rather not
                muck with the data at that stage.

                """
                if exch.value == 0 and self._linked:
                    continue
                self._print('## Exch %s [%s] (%g)' % (exch.flow, exch.direction, exch.value))

                term = exch.termination
                is_ref = exch.is_ref

                if is_ref:
                    if self._linked:  # in new regime, ref exch is already added
                        '''
                        if self._linked:
                            print('Squashing bad termination in linked reference exchange, %s\nFlow %s Term %s' % (
                                p.external_ref, exch.flow.external_ref, exch.termination))
                            term = None
                        '''
                        continue
                    else:
                        if exch.termination is not None:
                            print('Removing reference status from linked reference exchange, %s\nFlow %s Term %s' % (
                                p.external_ref, exch.flow.external_ref, exch.termination))
                            is_ref = False

                x = p.add_exchange(exch.flow, exch.direction, reference=rx, value=exch.value,
                                   termination=term)
                if len(exch.comment) > 0:
                    x.comment = exch.comment

                if not self._linked:
                    if is_ref:
                        self._print('## ## Exch is reference %s %s' % (exch.flow, exch.direction))
                        p.set_reference(exch.flow, exch.direction)
        return p

    def find_tag(self, process_uuid, rf_uuid, tag):
        return find_tag(self.objectify(process_uuid, rf_uuid), tag)

    '''
    def _fetch(self, uid, ref_flow=None):
        """
        ecospoldV2 files are named by activityId_referenceFlow - if none is supplied, take the first one found
        matching activityId
        :param uid:
        :param ref_flow:
        :return:
        """
        if ref_flow is not None:
            uid = '_'.join([uid, ref_flow])
        files = self.list_datasets(uid)
        if len(files) == 0:
            return None
        return self._create_process(files[0])
    '''

    def _fetch(self, ext_ref, **kwargs):
        """
        We want to handle two different kinds of external references: spold filenames (uuid_uuid.spold) and simple
        entity uuids.  In either case, we want to load *all* the files corresponding to the process, even if only
        one file is specified.  That means we want to truncate the key so that it is a maximum of one UUID long (but
        it would still be nice to allow the user to specify processes by partial uuid)
        :param ext_ref:
        :param kwargs:
        :return:
        """
        if ext_ref in self._master.elementary_exchanges:
            return self._create_flow(self._master.elementary_exchanges[ext_ref])

        if ext_ref in self._master.intermediate_exchanges:
            return self._create_flow(self._master.intermediate_exchanges[ext_ref])

        p_uuid, _ = spold_reference_flow(ext_ref)  # will error if at least one UUID is not present

        p = self[p_uuid]

        for r_uuid in self._process_flow_map[p_uuid]:
            p = self._create_process_and_single_reference(p_uuid, r_uuid, **kwargs)
        return p

    def retrieve_lcia_scores(self, process_uuid, rf_uuid, quantities=None):
        """
        This function retrieves LCIA scores from an Ecospold02 file and stores them as characterizations in
        an LcFlow entity corresponding to the *first* (and presumably, only) reference intermediate flow

        Only stores cfs for quantities that exist locally.
        :param process_uuid:
        :param rf_uuid:
        :param quantities: list of quantity entities to look for (defaults to all local lcia_methods)
        :return: a dict of quantity uuid to score
        """
        if quantities is None:
            quantities = [l for l in self.entities_by_type('quantity') if l.is_lcia_method]

        import time
        start_time = time.time()
        print('Loading LCIA results for %s_%s' % (process_uuid, rf_uuid))
        o = self.objectify(process_uuid, rf_uuid)

        self._print('%30.30s -- %5f' % ('Objectified', time.time() - start_time))
        p = self._create_process_entity(o)
        rf = self._grab_reference_flow(o, rf_uuid)

        exch = ExchangeValue(p, rf, 'Output', value=1.0)

        tags = dict()
        for q in quantities:
            if 'Method' in q.properties():
                if q['Name'] in tags:
                    raise KeyError('Name collision %s' % q['Name'])
                tags[q['Name']] = q

        results = LciaResults(p)

        for char in find_tag(o, 'flowData').getchildren():
            if 'impactIndicator' in char.tag:
                m = char.impactMethodName.text
                c = char.impactCategoryName.text
                i = char.name.text
                v = float(char.get('amount'))
                my_tag = ', '.join([m, c, i])
                if my_tag in tags:
                    q = tags[my_tag]
                    result = LciaResult(q)
                    cf = QRResult(rf['Name'], rf.reference_entity, q, rf.context, p['SpatialScope'], self.ref, v)
                    result.add_score(p.external_ref, exch, cf)
                    results[q.link] = result

        self._print('%30.30s -- %5f' % ('Impact scores collected', time.time() - start_time))

        return results

    def _load_all(self, exchanges=True, bailout=None):
        now = time()
        count = 0
        for p_u, r_set in self._process_flow_map.items():
            for r_u in r_set:
                self._create_process_and_single_reference(p_u, r_u, exchanges=exchanges)
            count += 1
            if count % 100 == 0:
                print(' Loaded %d processes (t=%.2f s)' % (count, time()-now))
            if bailout is not None:
                if count > bailout:
                    print('Bailing out...')
                    break

        print(' Loaded %d processes (t=%.2f s)' % (count, time() - now))
        self.check_counter()



from setuptools import setup, find_packages

VERSION = '0.1.8'

requires = [
    "synonym_dict>=0.2.0",
    "antelope_interface>=0.1.7",
    "xlstools>=0.1.0",
    "python-magic>=0.4.18",
    "requests>=2.25"
]

# optional: pylzma
"""
Version History
0.1.8   2022-04-08 - PyPI release to roll up a few small changes:
 * upgrade ecoinvent_lcia to use xlstools; port 3.8
 - pull out parse_math
 - windows path mgmt
 - background trivial impl needs to access entities and not refs
 - LciaResults can sum together if at least one has a unit node weight
 - cleanup reference quantities; finally add node activity as canonical quantity
 - quantity unit mismatch refactor


0.1.7.3 2021-08-11 - fix last edit
0.1.7.2 2021-08-11 - make lxml optional when running _find_providers
0.1.7.1 2021-08-11 - eliminate recursion error in basic get_reference()
0.1.7 - 2021-08-05 - merge configuration changes from [virtualize] in-progress
0.1.6 - 2021-03-10 - update to handle new synonym_dict 0.2.0, along with OLCA reference flow matching, ecoinvent 2.2,
                     a range of other improvements in performance and context handling
                     2021-03-10 post1
                     2021-03-17 post2 - get_context(); bump antelope version
                     
0.1.5 - 2021-02-05 - Updates to NullContext handling, flow term matching, fixed faulty requirements (add requests) 
0.1.4 - 2021/01/29 - background passing
0.1.3 - 2021/01/29 - Build passing (without bg tested)
0.1.3rc4           - fix last edit
0.1.3rc3           - xlrd removes support for xlsx in 2.0; _localize_file(None) return None
0.1.3rc2           - include data files
0.1.3rc1           - update requirements

0.1.2b - 2020/12/29 - fix some minor items
0.1.2 - 2020/12/28 - PyPI installation; includes significant performance enhancements for LCIA 

0.1.1 - 2020/11/12 - Bug fixes all over the place.  
                     Catalogs implemented
                     LCIA computation + flat LCIA computation reworked
                     Improvements for OpenLCA LCIA methods

0.1.0 - 2020/07/31 - Initial release - JIE paper
"""

setup(
    name="antelope_core",
    version=VERSION,
    author="Brandon Kuczenski",
    author_email="bkuczenski@ucsb.edu",
    license="BSD 3-Clause",
    install_requires=requires,
    extras_require={
        'XML': ['lxml>=1.2.0'],
        'write_to_excel': ['xlsxwriter>=1.3.7']
    },
    include_package_data=True,
    url="https://github.com/AntelopeLCA/core",
    summary="A reference implementation of the Antelope interface for accessing a variety of LCA data sources",
    long_description_content_type='text/markdown',
    long_description=open('README.md').read(),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Topic :: Scientific/Engineering"
    ],
    python_requires='>=3.6',
    packages=find_packages()
)



"""
Edelen and Ingwersen et al 2017:
Recommendations: "... separating flowable names from context and unit information ..."

Term Manager - used to handle string references to entities that are known by several synonyms.  Specifically:
 - contexts, which are drawn from flows' specified 'Compartment' or other classifying properties
 - flowables, which are drawn from flows' uuid and external link, and Name, CasNumber, Synonyms properties
 - quantities, which should be recognizable by external ref or uuid

The main objective of the TermManager is to enable the use of generalized synonym matching in performing LCIA. It
collects two kinds of mapping information: mapping string terms to flowables and contexts; and mapping (quantity,
flowable, context) tuples to [regionalized] characterization objects.

LciaEngine is designed to handle information from multiple origins to operate as a qdb in a catalog context.  The
subclass adds canonical lists of flowables and contexts that would be redundant if loaded into individual archives,
introduces "smart" hierarchical context lookup (CLookup), and adds the ability to quell biogenic CO2 emissions.
Someday, it might make sense to expose it as a massive, central graph db.
"""
from collections import namedtuple

from synonym_dict import SynonymDict, InconsistentLineage

from antelope import EntityNotFound
from ..contexts import ContextManager, Context, NullContext
from .quantity_manager import QuantityManager

from ..characterizations import Characterization, DuplicateCharacterizationError


DuplicateCharacterization = namedtuple('DuplicateCharacterization', ('qq_link', 'fb', 'cx_list', 'origin', 'flowable', 'location'))


class FactorCollision(Exception):
    pass


class QuantityConflict(Exception):
    pass


class NoFQEntry(Exception):
    """
    This exception has the specific meaning that there is no lookup for the named flow-quantity pair
    """
    pass


class OriginStrangeness(Exception):
    pass


from collections import defaultdict


class TermManager(object):
    """
    A TermManager is an archive-specific mapping of string terms to flowables and contexts.  During normal operation
    it is captive to an archive and automatically harvests flowable and context terms from flows when added to the
    archive.  It also hosts a set of CLookups for the archive's quantities, which enable fuzzy traversal of
    context hierarchies to find best-fit characterization factors.

    When a new entity is added to the archive, there are two connections to make: combine uuids and human-readable
    descriptors into a common database of synonyms, and connect those sets of synonyms to a set of entities known
    to the local archive.

    When harvesting terms from a new flow, the general approach is to merge the new entry with an existing entry
    if the existing one uniquely matches.  If there is more than one match, there are three general strategies:

      'prune': trim off the parts that match existing entries and make a new entry with only the distinct terms.
       this creates a larger more diffuse

      'merge': merge all the terms from the new entry and any existing entries together

    The mapping is prolific in both cases- adds a flowable-to-flow mapping for every flowable that matches a given flow
    (merge_strategy='merge' ensures that this is exactly one flowable), and adds the CF to every flowable that matches.

    USAGE MODEL: TODO
    The Term Manager is basically a giant pile of Characterization objects, grouped by canonical query quantity,
    flowable, and context. The basic TermManager enforces a restriction of one CF per qq | fb | cx combination.  The
    LciaEngine subclass uses a CLookup that can either be strict or non-strict, but nonetheless enforces one CF per
    qq | fb | cx | origin combination.

    There are thus four things the Term Manager must be able to do:
     - store names of flowables and contexts from source data
     - retrieve canonical flowable and context names from the store
     - store characterizations, properly indexed by canonical names
     - retrieve characterizations according to a query

    The components used to accomplish this are:
     _cm and _fm: the context and flowable synonym sets
     _flow_map: reverse-maps flowable terms to flows

     _q_dict: a 3-level nested dictionary:
         defaultdict: quantity uuid -> defaultdict: flowable -> CLookup: context-> {CFs}
       - first level defaultdict maps quantity uuid to second level
         - second level defaultdict maps flowable canonical name to CLookup / subclass
           - third level CLookup maps context to a set of CFs
     _fq_map: reverse-maps flowable canonical name to a set of quantities that characterize it
    """
    def __init__(self, contexts=None, flowables=None, quantities=None, merge_strategy='graft', quiet=True):
        """
        :param contexts: optional filename to initialize CompartmentManager
        :param flowables: optional filename to initialize FlowablesDict
        :param merge_strategy: (can also be specified at add_flow())

           'graft': - on conflict, follow incoming flow's link or name; add new terms to the existing flowable and
            discard conflicting terms.  solves many "shared CAS number" problems-- by depriving the new flowable of the
            CAS identifier

           'prune' or 'distinct': - on conflict, create a new flowable containing only new terms, discard all
            conflicting terms. This is the default for flows added with no context (so that distinct intermediate
            flows with the same name e.g. "bolt 10mm" could have different characterizations. If you want to
            prevent this, give your flow any nonempty context)

           'merge': - aggressively merge all co-synonymous flowables.  not tested.

        :param quiet:

        """
        # the synonym sets
        self._cm = ContextManager(source_file=contexts)
        self._fm = SynonymDict(source_file=flowables, entry_group="Flowables")
        self._qm = QuantityManager(source_file=quantities)

        self._q_dict = dict()  # dict of dicts. _q_dict[canonical quantity][canonical flowable] -> a context lookup
        #
        self._cl_typ = dict

        # the reverse mappings
        self._flow_map = defaultdict(set)  # maps flowable to flows having that flowable
        self._fq_map = dict()  # maps flowable to quantities characterized by that flowable

        # track duplicates
        self._dupes = []  #  List[DuplicateCharacterization]

        # config
        self._merge_strategy = merge_strategy
        self._quiet = bool(quiet)

    '''
    Utilities
    '''
    def is_context(self, item):
        try:
            self._cm.__getitem__(item)
            return True
        except KeyError:
            return False

    def __getitem__(self, item):
        """
        TermManager.__getitem__ retrieves a context known to the TermManager, or None if one is not found.
        Getitem exposes only the contexts, since flow external_refs are used as flowable synonyms
        :param item:
        :return:
        """
        if item is None:
            return None
        try:
            return self._cm.__getitem__(item)
        except KeyError:
            return None

    def _print(self, *args):
        if not self._quiet:
            print(*args)

    def add_terms(self, term_type, *terms, **kwargs):
        d = {'context': self._cm.new_entry,
             'flow': self._create_flowable,
             'flowable': self._create_flowable,
             'quantity': self._qm.new_entry}[term_type]
        d(*terms, **kwargs)

    @staticmethod
    def is_context(cx):
        return isinstance(cx, Context)

    @property
    def quiet(self):
        return self._quiet

    def add_quantity(self, quantity):
        if quantity.entity_type != 'quantity':
            raise TypeError('Must be quantity type')
        try:
            ex = self._qm[quantity.link]
            if not ex is quantity:
                raise QuantityConflict('Incoming %s does not match existing\n%s' % (ex, quantity))
        except KeyError:
            self._qm.add_quantity(quantity)
        return self._canonical_q(quantity)

    '''
    Info Storage
    def set_context(self, context_manager):
        """
        A flow will set its own context- but it needs a context manager to do so.

        Not sure whether to (a) remove FlowWithoutContext exception and test for None, or (b) allow set_context to
        abort silently if context is already set. Currently chose (b) because I think I still want the exception.


        :param context_manager:
        :return:
        """
        if context_manager.is_context(self._context):
            # cannot change context once it's set
            return
        if self.has_property('Compartment'):
            _c = context_manager.add_compartments(self['Compartment'])
        elif self.has_property('Category'):
            _c = context_manager.add_compartments(self['Category'])
        else:
            _c = context_manager.get('none')
            # raise AttributeError('Flow has no contextual attribute! %s' % self)
        if not context_manager.is_context(_c):
            raise TypeError('Context manager did not return a context! %s (%s)' % (_c, type(_c)))
        self._context = _c
        self._flowable = context_manager.add_flow(self)
    '''
    def _context_and_parents(self, cx):
        c = self[cx]
        while c is not None:
            yield c
            c = c.parent

    def _add_compartments(self, comps):
        """
        Complicated issue here-- basically, we have to decide on a conflict resolution strategy for
        ContextManager.add_compartments() -- and that strategy must be different for TermManager and LciaEngine.
        For TermManager, the most important thing is to not mess up the literal context specifications, because
        fidelity to source is paramount. Therefore, we should 'rename' contexts and keep them separated.

        Whereas, in LciaEngine the main concern is to harmonize contexts, which means forcing them to play well
        together. This is why we have hints in LciaEngine but not in TermManager.  That is also why the default is to
        attach trees together, rather than duplicate them.

        To TEST: try adding ('emissions', 'emissions to water') and ('elementary flows', 'emissions to water')-- the
        outcome should be different for TermManager and LciaEngine.
        :param comps:
        :return:
        """
        return self._cm.add_compartments(comps, conflict='rename')

    def add_context(self, context, *terms, origin=None):
        """
        We are using conflict='attach' by default so that 'emissions' and 'resources' can be attached to
        'Elementary flows' if it exists. It may cause undesirable side-effects if different intermediate categories
        use similar terms, e.g. ('buildings', 'heat', 'steam') and ('chemical', 'heat', 'process') would result in
        either

        :param context:
        :param origin: apply origin to context if it has none
        :return:
        """
        if isinstance(context, str):
            context = (context,)
        cx = self._add_compartments(tuple(context))
        if origin is not None:
            if cx.fullname == cx.name:
                cx.add_origin(origin)
        for term in terms:
            self._cm.add_synonym(cx.name, term)
        return cx

    def _check_context(self, flow):
        """
        The point of this function is simply to log which local context corresponds to the flow's presented context
        :param flow:
        :return:
        """
        if not hasattr(flow, 'context'):
            return NullContext
        try:
            _c = self._cm[flow.get_context()]
        except (KeyError, InconsistentLineage):
            _c = self._add_compartments(flow.context)
        _c.add_origin(flow.origin)
        return _c

    def _create_flowable(self, name, *syns, prune=False):
        fb = self._fm.new_entry(name, *syns, prune=prune).object
        if fb in self._fq_map:
            self._print('Adding terms to existing flowable %s' % fb)
        else:
            try:
                self._print('Adding new flowable %s' % fb)
            except TypeError:
                print('Failed to create flowable for %s (%s)' % (name, syns))
                raise
            self._fq_map[fb] = set()
        return fb

    def _merge_terms(self, dominant, *syns):
        """
        Two parts to this: merge the entries in the flowables manager; update local reverse mappings:
         _q_dict
         _fq_map (nonconflicting)
         _flow_map (nonconflicting)
        Before we can do either, we need to check for collisions
        This just brute forces it which must be ungodly slow, and btw it also hasn't been tested
        :param dominant:
        :param syns:
        :return:
        """
        fq_conflicts = []
        for f_dict in self._q_dict.values():
            try:
                combo = f_dict[dominant]
            except KeyError:
                combo = self._cl_typ()
            for syn in syns:
                if syn not in f_dict:
                    continue
                cl = f_dict[syn]
                for k in cl.keys():
                    if k in combo:
                        cand = (combo[k], cl[k])
                        if cand[0].value == cand[1].value:
                            continue
                        fq_conflicts.append(cand)  # but it's really only a conflict if the cf values differ

        if len(fq_conflicts) > 0:
            print('%d Merge conflicts encountered' % len(fq_conflicts))
            return fq_conflicts

        for f_dict in self._q_dict.values():
            try:
                combo = f_dict[dominant]
            except KeyError:
                combo = self._cl_typ()
            for syn in syns:
                if syn not in f_dict:
                    continue
                combo.update(f_dict[syn])
            f_dict[dominant] = combo

        if dominant not in self._fq_map:
            self._fq_map[dominant] = set()
        for syn in syns:
            self._fq_map[dominant] += self._fq_map.pop(syn, set())
            self._flow_map[dominant] += self._flow_map.pop(syn, set())

        for syn in syns:
            self._fm.merge(dominant, syn)
        return dominant

    @staticmethod
    def _flow_terms(flow):
        """
        This function was created because we don't want TermManagers getting confused about things like misassigned
        CAS numbers, because they don't have the capacity to store multiple CFs.  So we will save the full synonym
        list for the LciaEngine.
        :param flow:
        :return:
        """
        return flow.name, flow.link

    def flows_for_flowable(self, fb):
        if fb in self._flow_map:
            for f in self._flow_map[fb]:
                yield f

    def _add_to_existing_flowable(self, fb, new_terms):
        for term in new_terms:
            self._fm.add_synonym(str(fb), term)

    def _add_flow_terms(self, flow, merge_strategy=None):
        """
        This process takes in an inbound FlowInterface instance, identifies the flowable(s) that match its terms, and
        adds new terms to the existing or new flowable.  May update a flow's name in case of conflict.
        :param flow:
        :param merge_strategy:
        :return:
        """
        merge_strategy = merge_strategy or self._merge_strategy
        fb_map = defaultdict(list)  # list instead of set because sequence matters
        for syn in self._flow_terms(flow):  # we rely on SynonymSet.terms yielding flow.name first
            # make a list of all the existing flowables that match the incoming flow
            fb_map[self._fm.get(syn)].append(syn)
        new_terms = fb_map.pop(None, [])

        if len(fb_map) == 0:  # all new terms
            if len(new_terms) == 0:
                raise AttributeError('Flow appears to have no terms: %s' % flow)
            fb = self._create_flowable(flow.name)

        elif merge_strategy in ('prune', 'distinct'):
            # fb = self._add_distinct_terms(flow, new_terms)
            try:
                fb = self._create_flowable(new_terms.pop(0), prune=True)  # new_terms cannot be 0
            except IndexError:
                fb = self._fm.get(flow.link)  # um, I guess if there are no new terms...

        elif len(fb_map) == 1:  # one existing match- graft onto that
            fb = list(fb_map.keys())[0]

        else:  # > 2 matches-- invoke merge strategy
            if merge_strategy == 'graft':
                """
                We want link to be dominant but name to be preferred. we have 3 cases:
                 - link is new, name is new: all new flowable with new terms
                 - link is new, name is old: add new terms to old flowable
                 - link is old: add only new terms to old flowable, including or excluding name
                 
                
                note that the keys of fb_map are flowables and the values are incoming terms. but anything remaining 
                in fb_map at this point is a known term so does not need to be added
                """
                self._print('\nPruning entry for %s' % flow)
                if flow.link in new_terms:
                    if flow.name in new_terms:
                        fb = self._create_flowable(flow.name)
                    else:
                        fb = self._fm[flow.name]
                else:
                    fb = self._fm[flow.link]

            elif merge_strategy == 'merge':
                # this is trivial but has never been tested, I mean even once
                self._print('Merging')
                fb = self._merge_terms(*fb_map.keys())
                if isinstance(fb, list):
                    raise FactorCollision(fb)
            else:
                raise ValueError('merge strategy %s' % self._merge_strategy)

        self._add_to_existing_flowable(fb, new_terms)

        # validate that the flow's name looks up proper flowable
        if self._fm[flow.name] is not fb:
            try:
                flow.name = next(k for k in self._flow_terms(flow) if self._fm[k] is fb)
            except StopIteration:
                raise AttributeError('None of the synonyms map to flowable [%s]: %s' % (fb, flow))

        # log the flowables that match the flow
        self._flow_map[fb].add(flow)
        for _tf in fb_map.keys():
            self._flow_map[_tf].add(flow)  # all this belongs in a Graph DB
        return fb

    def add_flow(self, flow, merge_strategy=None):
        """
        We take a flow from outside and add its known terms. That means
         - adding flow's reference quantity
         - merging any context with the local context tree;
         - adding flowable terms to the flowables list;
         - mapping flow to all flowables (assigning flowable if none is found)

        Note that all flows having null context are forced to be distinct from one another. If you want two flowables
        to share properties, you must add them having non-null contexts. not sure how I feel about this but it does work.

        :param flow:
        :param merge_strategy: overrule default merge strategy
        :return: the Flowable object to which the flow's terms have been added
        """
        if hasattr(flow, 'reference_entity') and flow.reference_entity is not None:
            if isinstance(flow.reference_entity, str):
                try:
                    self.get_canonical(flow.reference_entity)
                except EntityNotFound:
                    raise ValueError('unrecognized quantity spec %s ' % flow.reference_entity)
            else:
                self.add_quantity(flow.reference_entity)  # ensure exists
        cx = self._check_context(flow)
        if cx is NullContext:
            merge_strategy = 'distinct'  # keep distinct terms for null-context flows
        return self._add_flow_terms(flow, merge_strategy=merge_strategy)

    def add_characterization(self, flowable, ref_quantity, query_quantity, value, context=None, origin=None,
                             location=None, overwrite=False):
        """
        Replacement for flow-based add_characterization.  THE ONLY place to create Characterization objects.
        Add them to all flowables that match the supplied flow.
        :param flowable: if not known to the flowables dict, added as a new flowable
        :param ref_quantity: [entity or ref]
        :param query_quantity: [entity or ref]
        :param value: mandatory. either a numeric value or a dict mapping locations to values
        :param context: the context for which the characterization applies.  Should be a string or tuple.  None means
          the factor applies to all contexts.
        :param overwrite: whether to overwrite an existing value if it already exists (ignored if value is a dict)
        :param location: (ignored if value is a dict) 'GLO' used if no location is provided
        :param origin: (optional; origin of value; defaults to quantity.origin)
        :return: created or updated characterization
        """
        if origin is None:
            origin = query_quantity.origin

        try:
            cx = self._cm[context]
        except KeyError:
            cx = self.add_context(context, origin=origin)

        try:
            fb = self._fm[flowable]
        except KeyError:
            fb = self._create_flowable(flowable)

        try:
            rq = self._canonical_q(ref_quantity)
        except KeyError:
            rq = self.add_quantity(ref_quantity)

        try:
            qq = self._canonical_q(query_quantity)
        except KeyError:
            qq = self.add_quantity(query_quantity)

        cf = self._find_exact_cf(qq, fb, cx, origin, flowable)

        if location is None:
            location = 'GLO'

        if cf is None:
            # create our new Characterization with the provided flowable, not the detected flowable
            new_cf = Characterization(flowable, rq, qq, context=cx, origin=origin)
            if isinstance(value, dict):
                new_cf.update_values(**value)
            else:
                if value == 0:
                    self._print('zero CF: %s, %s, %s, %s, %s, %s' % (flowable, rq.link, qq.link, cx, origin, location))
                new_cf.add_value(value, location=location)

            # add our new CF to the lookup tree
            self._qassign(qq, fb, new_cf, context=cx)

            return new_cf
        else:
            if cf.ref_quantity != rq:
                # create a conversion-factor instead:
                if value == 0:
                    # we cannot interpret a 0 as a cf- so we ditch
                    return cf

                try:
                    factor = value / cf[location]
                except ZeroDivisionError:
                    try:
                        factor = value / cf.value
                    except TypeError:  # dict = fail
                        factor = value

                # this recurses to add a cf between our flow's ref quantity and the ref quantity of the already-seen cf
                try:
                    self.add_characterization(fb, rq, cf.ref_quantity, factor, context=cx, origin=origin, location=location)
                except DuplicateCharacterizationError as e:
                    print((qq.link, fb, cx.as_list(), origin, flowable, location))
                    print('recursing to %s %s' % (cf.ref_quantity.name, cf.ref_quantity.uuid))
                    print('ignoring duplicate characterization %s' % e)
                return cf
            # update entry in the lookup tree
            elif isinstance(value, dict):
                cf.update_values(**value)
            else:
                try:
                    cf.add_value(value, location=location, overwrite=overwrite)
                except DuplicateCharacterizationError as e:
                    self._dupes.append(DuplicateCharacterization(qq.link, fb, cx.as_list(), origin, flowable, location))
                    print('ignoring duplicate characterization %s' % e)
            return cf

    '''
    Info Retrieval
    '''
    def _find_exact_cf(self, qq, fb, cx, origin, flowable):
        """
        The purpose of this function is to retrieve an exact CF if one exists.
        WHY does this not take rq into account??

        origin used in subclasses
        :param qq:
        :param fb:
        :param cx:
        :param origin:
        :param flowable: using this in subclass
        :return:
        """
        try:
            clookup = self._qlookup(qq, fb)
        except NoFQEntry:
            return None
        if cx in clookup:
            return clookup[cx]
        return None

    def get_canonical(self, quantity):
        try:
            return self._canonical_q(quantity)
        except KeyError:
            raise EntityNotFound(quantity)

    def _canonical_q(self, quantity):
        """
        override here
        :param quantity: a quantity entity or descriptor
        :return: a canonical quantity or raise KeyError
        """
        try:
            return self._qm[quantity]
        except KeyError:
            if hasattr(quantity, 'link'):
                try:
                    return self._qm[quantity.link]
                except KeyError:
                    return self._qm.find_matching_quantity(quantity)
            return self._qm[str(quantity)]

    def _canonical_q_ref(self, quantity):
        return self._canonical_q(quantity).external_ref

    def _qlookup(self, qq, fb):
        """
        Returns a mapping from context to CF
        :param qq: a canonical quantity
        :param fb: a canonical flowable
        :return:
        """
        try:
            return self._q_dict[qq][fb]
        except KeyError:
            raise NoFQEntry

    def _qaccess(self, qq, fb=None):
        """
        creates an entry if none exists
        :param qq: a canonical quantity
        :param fb [None]: a canonical flowable. If None, return the quantity dict
        :return:
        """
        if qq in self._q_dict:
            qd = self._q_dict[qq]
        else:
            qd = dict()
            self._q_dict[qq] = qd
        if fb is None:
            return qd
        elif fb in qd:
            return qd[fb]
        else:
            cl = self._cl_typ()
            qd[fb] = cl
            return cl

    @staticmethod
    def _store_cf(cl, context, new_cf):
        """
        Assigns the cf to the mapping; does subclass-specific collision checking
        :param cl:
        :param new_cf:
        :return:
        """
        if context in cl:
            raise FactorCollision
        cl[context] = new_cf

    def _qassign(self, qq, fb, new_cf, context=None):
        """
        Assigns the new_cf to the canonical quantity and flowable, taking context and origin from new_cf
        :param qq: a canonical quantity
        :param fb: a canonical flowable
        :param new_cf: a characterization, having a canonical context
        :return:
        """
        if context is None:
            context = new_cf.context
        cl = self._qaccess(qq, fb)
        self._store_cf(cl, context, new_cf)
        self._fq_map[fb].add(qq)

    def _factors_for_flowable(self, fb, qq, cx, **kwargs):
        """
        detach lookup for cleanness. canonical everything
        :param fb:
        :param qq:
        :param cx:
        :param kwargs: used in subclasses
        :return:
        """
        try:
            cl = self._qlookup(qq, fb)
        except NoFQEntry:
            return
        if cx is None:
            for v in cl.values():
                yield v
        else:
            if cx in cl:
                yield cl[cx]

    def factors_for_flowable(self, flowable, quantity=None, context=None, **kwargs):
        """
        This is the method that actually performs the lookup.  Other methods are wrappers for this.

        If context is None, this is "unspecified". It matches all CFs regardless of context.

        Core to this is getting a canonical context, which is done by __getitem__.  In TermManager, this returns
        None for contexts that are not known
        :param flowable: a string
        :param quantity: a quantity known to the quantity manager
        :param context: [None] default provide all contexts; must explicitly provide 'none' to filter by null context
        :return:
        """
        try:
            fb = self._fm[flowable]
            cx = self[context]
        except KeyError:
            return
        if quantity is None:
            for qq in self._fq_map[fb]:
                for cf in self._factors_for_flowable(fb, qq, cx, **kwargs):
                    yield cf
        else:
            qq = self._canonical_q(quantity)
            for cf in self._factors_for_flowable(fb, qq, cx, **kwargs):
                yield cf

    def factors_for_quantity(self, quantity, flowable=None, context=None, **kwargs):
        """
        param dist [0] only used if compartment is specified. by default report only exact matches.

        :param quantity:
        :param flowable:
        :param context:
        :return:
        """
        if flowable is not None:
            for k in self.factors_for_flowable(flowable, quantity=quantity, context=context, **kwargs):
                yield k
        else:
            qq = self._canonical_q(quantity)
            for f in self._qaccess(qq).keys():
                for k in self.factors_for_flowable(f, quantity=qq, context=context, **kwargs):
                    yield k

    def get_flowable(self, term):
        """
        Input is a Flow or str
        :param term:
        :return:
        """
        if hasattr(term, 'synonyms'):
            terms = list(term.synonyms)
        elif hasattr(term, 'name'):
            terms = [term.name]
        else:
            terms = [str(term)]
        for t in terms:
            try:
                return self._fm[t]
            except KeyError:
                continue
        raise KeyError(term)

    def flowables(self, search=None, origin=None, quantity=None):
        """
        :param origin: used in subclass
        :param search:
        :param quantity:
        :return:
        """
        if quantity is None:
            qq = None
        else:
            qq = self.get_canonical(quantity)
        if search is None:
            if quantity is None:
                for fb in self._fm.objects:
                    yield fb
            else:
                for fb in self._q_dict[qq].keys():
                    yield fb
        else:
            for fb in self._fm.objects_with_string(search):
                if qq is None:
                    yield fb
                else:
                    if fb in self._q_dict[qq]:
                        yield fb

    def unmatched_flowables(self, flowables):
        """
        Given an iterable of flowable strings, return a list of entries that were not recognized as synonyms to known
        flowables
        :param flowables:
        :return:
        """
        unknown = []
        for fb in flowables:
            try:
                self._fm[fb]
            except KeyError:
                unknown.append(fb)
        return unknown

    def synonyms(self, term):
        """
        Search for synonyms, first in contexts, then flowables, then quantities.
        The somewhat awkward structure here is because of the dynamics of returning generators-- using
        try: return self._cm.synonyms(term) except KeyError: ... the KeyError was not getting caught because the
        generator was already returned before iterating.

        contexts are searched first, then quantities, then flowables
        :param term:
        :return:
        """
        try:
            obj = self._cm[term]
            it = self._cm.synonyms(obj)
        except KeyError:
            try:
                obj = self._qm[term]
                it = self._qm.synonyms(obj)
            except KeyError:
                try:
                    obj = self._fm[term]
                    it = self._fm.synonyms(obj)
                except KeyError:
                    it = ()
        for k in it:
            yield k

    def contexts(self, search=None, origin=None):
        for cx in self._cm.objects:
            if origin is not None:
                if not cx.has_origin(origin):
                    continue
            if search is None:
                yield cx
            else:
                if cx.contains_string(search):
                    yield cx

    def quantities(self, search=None, origin=None):
        for q in self._qm.objects:
            if origin is not None:
                if not q.origin.startswith(origin):
                    continue
            if search is None:
                yield q
            else:
                if q.contains_string(search):
                    yield q

    '''
    De/Serialization
    
    Strategy here is that term manager serializations should be totally self-contained and also minimal. 
    
    Well, not totally self-contained, since quantities themselves are not going to be messed with.
    
    So aside from quantities, which are not going to be messed with, the term manager serialization should be specified
    by quantity. The routine must determine, for all query quantities listed:
     (a) the set of reference quantities encountered
     (b) the set of flowables encountered
     (c) the set of contexts encountered
    The reference and query quantities must all be serialized per normal, with others omitted.  The term manager 
    section should include flowables, contexts, and characterizations, but only the minimal ones required to cover the
    specified quantities.
    
    Then de-serialization is straightforward: after quantities are loaded, simply apply all flowables, contexts, and
    then characterizations- flows not required!
    '''
    def _serialize_qdict(self, origin, quantity, values=False):
        _ql = self._qaccess(quantity)
        d = {}
        for fb, cl in _ql.items():
            _od = {str(c): cf.serialize(values=values, concise=True) for c, cf in cl.items() if cf.origin == origin}
            if len(_od) > 0:
                d[str(fb)] = _od
        return d

    def _serialize_factors(self, origin, *quantities, values=False):
        """
        This does not exactly work because, if done from an lcia-engine the result will not be closed to the origin
        specified.  Anyway, in the single-origin case this should be the default.

        in LciaEngine We [may] need a way to record the original flowable + context names, so that origin-specific
        serialization stays closed.
        :param origin:
        :param quantities:
        :param values:
        :return:
        """
        if len(quantities) == 0:
            qqs = self._q_dict.keys()
        else:
            qqs = [self._canonical_q(q) for q in quantities]
        j = dict()
        for q in qqs:
            _sq = self._serialize_qdict(origin, q, values=values)
            if len(_sq) > 0:
                j[self._canonical_q_ref(q)] = _sq
        return j

    def serialize(self, origin, *quantities, values=False):
        """

        :param origin:
        :param quantities:
        :param values:
        :return: 3-tuple:
         - serialized term manager as dict,
         - set of query quantity external refs,
         - set of reference quantity uuids
           == not sure WHY UUIDs rather than external refs, but that is what characterizations.py gives us
        """
        qqs = set()  # query quantities to serialize
        rqs = set()  # reference quantities to serialize
        fbs = set()  # flowables to serialize
        cxs = set()  # contexts to serialize

        # serialize factors themselves
        j = {'Characterizations': self._serialize_factors(origin, *quantities, values=values)}

        # identify terms + qs
        for qq, factors in j['Characterizations'].items():
            qqs.add(qq)
            for fb, char in factors.items():
                fbs.add(fb)
                for cx, cf in char.items():
                    for c in self._context_and_parents(cx):
                        cxs.add(str(c))
                    try:
                        rqs.add(cf['ref_quantity'])
                    except KeyError:
                        print('%s\n%s' % (cx, cf))
                        raise

        # add flowables and contexts
        j.update(self._fm.serialize(obj for obj in self._fm.entries if str(obj) in fbs))
        j.update(self._cm.serialize(obj for obj in self._cm.entries if str(obj) in cxs))

        return j, qqs, rqs

    def add_from_json(self, j, q_map, origin=None):
        """

        :param j:
        :param q_map: a dict whose keys are external_refs and uuids, and whose values are quantities
        :param origin:
        :return:
        """
        if 'SynonymSets' in j and 'Flowables' not in j:
            j['Flowables'] = j.pop('SynonymSets')

        self._cm.load_dict(j)  # automatically pulls out 'Compartments'
        self._fm.load_dict(j)
        for f in self._fm.objects:
            if f not in self._fq_map:
                self._fq_map[f] = set()
        self._add_from_json(j['Characterizations'], q_map, origin)

    def _add_from_json(self, j, q_map, origin):
        """
        Argument: the contents of archive['characterizations'], which looks like this (general case):
        'termManager': {
          'SynonymSets': [ {flowable}..],
          'Compartments': [ {context}..],
          'Characterizations': {
          query_quantity.external_ref: {
            flowable: {
              context: {
                origin: {
                  ref_quantity: xxx,
                  value: {
                    locale: val
                  }
                }
              }
            }
          }
        }

        (single origin case):
        'characterizations': {
          'SynonymSets': [ {flowable}..],
          'Compartments': [ {context}..],
          query_quantity.external_ref: {
            flowable: {
              context: {
                ref_quantity: xxx,
                value: {
                  locale: val
                }
              }
            }
          }
        }

        :param j:
        :return:
        """
        for query_ext_ref, fbs in j.items():
            try:
                query_q = q_map[query_ext_ref]
            except KeyError:
                query_q = self._canonical_q(query_ext_ref)
            for fb, cxs in fbs.items():
                if origin is None:
                    for cx, cfs in cxs.items():
                        for org, spec in cfs.items():
                            try:
                                ref_q = q_map[spec['ref_quantity']]
                            except KeyError:
                                ref_q = self._canonical_q(spec['ref_quantity'])
                            self.add_characterization(fb, ref_q, query_q, spec['value'], context=cx,
                                                      origin=org)
                else:
                    for cx, spec in cxs.items():
                        try:
                            ref_q = q_map[spec['ref_quantity']]
                        except KeyError:
                            ref_q = self._canonical_q(spec['ref_quantity'])
                        self.add_characterization(fb, ref_q, query_q, spec['value'], context=cx,
                                                  origin=origin)



"""
Query Interface -- used to operate catalog refs
"""

from antelope import (IndexInterface, BackgroundInterface, ExchangeInterface, QuantityInterface, EntityNotFound, UnknownOrigin)
#                      ForegroundInterface,
#                      IndexRequired, PropertyExists,
#                      )
from antelope.refs.exchange_ref import RxRef

INTERFACE_TYPES = {'basic', 'index', 'exchange', 'background', 'quantity', 'foreground'}
READONLY_INTERFACE_TYPES = {'basic', 'index', 'exchange', 'background', 'quantity'}

def zap_inventory(interface, warn=False):
    if interface == 'inventory':
        if warn:
            print('# # # # # # # # # **** Warning: use exchange over inventory ***** # # # # # # # # #')
            raise AttributeError
        return 'exchange'
    return interface


class NoCatalog(Exception):
    pass


class BackgroundSetup(Exception):
    pass


class BadInterfaceSpec(Exception):
    pass


class CatalogQuery(IndexInterface, BackgroundInterface, ExchangeInterface, QuantityInterface): # , ForegroundInterface):
    """
    A CatalogQuery is a class that performs any supported query against a supplied catalog.
    Supported queries are defined in the lcatools.interfaces, which are all abstract.
    Implementations also subclass the abstract classes.

    This reduces code duplication (all the catalog needs to do is provide interfaces) and ensures consistent signatures.

    The arguments to a query should always be text strings, not entities.  When in doubt, use the external_ref.

    The EXCEPTION is the bg_lcia routine, which works best (auto-loads characterization factors) if the query quantity
    is a catalog ref.

    The catalog's resolver performs fuzzy matching, meaning that a generic query (such as 'local.ecoinvent') will return
    both exact resources and resources with greater semantic specificity (such as 'local.ecoinvent.3.2.apos').
    All queries accept the "strict=" keyword: set to True to only accept exact matches.
    """
    _recursing = False
    _dbg = False

    def on_debug(self):
        self._dbg = True

    def off_debug(self):
        self._dbg = False

    def _debug(self, *args):
        if self._dbg:
            print(*args)

    def __init__(self, origin, catalog=None, debug=False):
        self._origin = origin
        self._catalog = catalog
        self._dbg = debug

        self._entity_cache = dict()
        self._iface_cache = dict()

    @property
    def origin(self):
        return self._origin

    @property
    def _tm(self):
        return self._catalog.lcia_engine

    def is_elementary(self, context):
        """
        Stopgap used to expose access to a catalog's Qdb; in the future, flows will no longer exist and is_elementary
        will be a trivial function of an exchange asking whether its termination is a context or not.
        :param context:
        :return: bool
        """
        return self._tm[context.fullname].elementary


    def cascade(self, origin):
        """
        Generate a new query for the specified origin.
        Enables the query to follow the origins of foreign objects found locally.
        :param origin:
        :return:
        """
        return self._grounded_query(origin)

    def _grounded_query(self, origin):
        if origin is None or origin == self._origin:
            return self
        return self._catalog.query(origin)

    def __str__(self):
        return '%s for %s (catalog: %s)' % (self.__class__.__name__, self.origin, self._catalog.root)

    def _iface(self, itype, strict=False):
        self._debug('Origin: %s' % self.origin)
        if self._catalog is None:
            raise NoCatalog
        if itype in self._iface_cache:
            self._debug('Returning cached iface')
            yield self._iface_cache[itype]
        for i in self._catalog.gen_interfaces(self._origin, itype, strict=strict):
            if itype == 'background':  # all our background implementations must provide setup_bm(query)
                self._debug('Setting up background interface')
                try:
                    i.setup_bm(self)
                except AttributeError:
                    raise BackgroundSetup('Failed to configure background')

            self._debug('yielding %s' % i)
            self._iface_cache[itype] = i  # only cache the most recent iface
            yield i

    def _perform_query(self, itype, attrname, exc, *args, strict=False, **kwargs):
        if itype is None:
            raise BadInterfaceSpec(itype, attrname)  # itype = 'basic'  # fetch, get properties, uuid, reference

        self._debug('Performing %s query, iface %s' % (attrname, itype))
        try:
            for iface in self._iface(itype, strict=strict):
                try:
                    self._debug('Attempting %s query on iface %s' % (attrname, iface))
                    result = getattr(iface, attrname)(*args, **kwargs)
                except exc:  # allow nonimplementations to pass silently
                    continue
                if result is not None:  #successful query must return something
                    return result
        except NotImplementedError:
            pass

        raise exc('itype %s required for attribute %s | %s' % (itype, attrname, args))

    def resolve(self, itype=INTERFACE_TYPES, strict=False):
        """
        Secure access to all known resources but do not answer any query
        :param itype: default: all interfaces
        :param strict: [False]
        :return:
        """
        for k in self._iface(itype, strict=strict):
            yield k

    def get(self, eid, **kwargs):
        """
        Retrieve entity by external Id. This will take any interface and should keep trying until it finds a match.
        :param eid: an external Id
        :return:
        """
        if eid not in self._entity_cache:
            entity = self._perform_query('basic', 'get', EntityNotFound, eid,
                                         **kwargs)
            self._entity_cache[eid] = self.make_ref(entity)
        return self._entity_cache[eid]

    def get_reference(self, external_ref):
        k = (external_ref, True)
        if k not in self._entity_cache:
            ref = self._perform_query('basic', 'get_reference', EntityNotFound, external_ref)
            # quantity: unit
            # flow: quantity
            # process: list
            # fragment: fragment
            #[context: context]
            if ref is None:
                deref = None
            elif isinstance(ref, list):
                deref = [RxRef(self.make_ref(x.process), self.make_ref(x.flow), x.direction, x.comment) for x in ref]
            elif ref.entity_type == 'unit':
                deref = ref.unitstring
            else:
                deref = self.make_ref(ref)
            self._entity_cache[k] = deref
        return self._entity_cache[k]

    '''
    LCIA Support
    get_canonical(quantity)
    catch get_canonical calls to return the query from the local Qdb; fetch if absent and load its characterizations
    (using super ==> _perform_query)
    '''
    def get_canonical(self, quantity, **kwargs):
        try:
            # print('Gone canonical')
            q_can = self._tm.get_canonical(quantity)
        except EntityNotFound:
            if hasattr(quantity, 'entity_type') and quantity.entity_type == 'quantity':
                print('Missing canonical quantity-- adding to LciaDb')
                self._catalog.register_quantity_ref(quantity)
                q_can = self._tm.get_canonical(quantity)
                # print('Retrieving canonical %s' % q_can)
            else:
                raise
        return q_can

    def make_ref(self, entity):
        if isinstance(entity, list):
            return [self.make_ref(k) for k in entity]
        if entity.entity_type == 'fragment':
            # TODO: create a new ForegroundQuery to eliminate the need for this hack
            return entity  # don't make references for fragments just now
        if entity is None:
            return None
        if entity.is_entity:
            try:
                e_ref = entity.make_ref(self._grounded_query(entity.origin))
            except UnknownOrigin:
                e_ref = entity.make_ref(self)
        else:
            e_ref = entity  # already a ref
        if entity.entity_type == 'quantity':
            # print('Going canonical')
            return self.get_canonical(e_ref)
        return e_ref



import os
import re
import gzip as gz
import json


def from_json(fname):
    """
    Routine to extract the contents of a json file.  no longer pretends to support PY2.
    :param fname: json file, optionally gzipped
    :return: a json-derived dict
    """
    print('Loading JSON data from %s:' % fname)
    if bool(re.search('\.gz$', fname)):
        with gz.open(fname, 'rt') as fp:
            j = json.load(fp)
    else:
        with open(fname, 'r') as fp:
            j = json.load(fp)
    return j


def to_json(obj, fname, gzip=False):
    dirname = os.path.dirname(fname)
    if not os.path.exists(dirname):
        os.makedirs(dirname, exist_ok=True)
    if gzip is True:
        if not bool(re.search('\.gz$', fname)):
            fname += '.gz'
        try:  # python3
            with gz.open(fname, 'wt') as fp:
                json.dump(obj, fp, indent=2, sort_keys=True)
        except ValueError:  # python2
            with gz.open(fname, 'w') as fp:
                json.dump(obj, fp, indent=2, sort_keys=True)
    else:
        with open(fname, 'w') as fp:
            json.dump(obj, fp, indent=2, sort_keys=True)



"""
Interim classes with useful building blocks
"""

from __future__ import print_function, unicode_literals

from antelope import LcQuery

from ..entities import LcProcess, ZeroAllocation
from ..implementations import ExchangeImplementation, BackgroundImplementation, LcConfigureImplementation
from .basic_archive import BasicArchive, BASIC_ENTITY_TYPES


LC_ENTITY_TYPES = BASIC_ENTITY_TYPES + ('process', )


class LcArchive(BasicArchive):
    """
    A class meant for storing and managing LCA data collections.  Adds processes as a supported entity type (contrast
    with LcForeground which adds fragments).

    To support processes, adds inventory, background, and configure interfaces.
    """
    _entity_types = LC_ENTITY_TYPES

    @property
    def query(self):
        return LcQuery(self)

    def make_interface(self, iface):
        if iface == 'inventory' or iface == 'exchange':
            return ExchangeImplementation(self)
        elif iface == 'background':
            return BackgroundImplementation(self)
        elif iface == 'configure':
            return LcConfigureImplementation(self)
        else:
            return super(LcArchive, self).make_interface(iface)

    def _add_children(self, entity):
        if entity.entity_type == 'process':
            # need to import all the process's flows
            for x in entity.exchanges():
                self.add_entity_and_children(x.flow)
        else:
            super(LcArchive, self)._add_children(entity)

    @staticmethod
    def _lcia_key(quantity):
        return ', '.join([quantity['Method'], quantity['Category'], quantity['Indicator']])

    def load_from_dict(self, j, _check=True, jsonfile=None):
        """
        Archives loaded from JSON files are considered static.
        :param j:
        :param _check:
        :return:
        """
        super(LcArchive, self).load_from_dict(j, _check=False, jsonfile=jsonfile)
        if 'processes' in j:
            for e in j['processes']:
                self.entity_from_json(e)
        if _check:
            self.check_counter()
        if jsonfile is not None and jsonfile == self.source:
            self._loaded = True

    def _process_from_json(self, entity_j, ext_ref):
        # note-- we are officially abandoning referenceExchange notation
        if 'referenceExchange' in entity_j:
            entity_j.pop('referenceExchange')
        a_b_q = entity_j.pop('AllocatedByQuantity', None)
        exchs = entity_j.pop('exchanges', [])
        process = LcProcess(ext_ref, **entity_j)
        refs, nonrefs = [], []
        ref_x = dict()
        for i, x in enumerate(exchs):
            if 'isReference' in x and x['isReference'] is True:
                refs.append(i)
            else:
                nonrefs.append(i)
        # first add reference exchanges
        for i in refs:
            x = exchs[i]
            # eventually move this to an exchange classmethod - which is why I'm repeating myself for now
            v = None
            f = self._get_entity(x['flow'])
            d = x['direction']
            if 'value' in x:
                v = x['value']
            ref_x[x['flow']] = process.add_exchange(f, d, value=v)
            process.set_reference(f, d)
        # then add ordinary [allocated] exchanges
        for i in nonrefs:
            x = exchs[i]
            # is_ref = False
            f = self._get_entity(x['flow'])
            d = x['direction']
            if 'termination' in x:
                t = x['termination']
                cx = self.tm[t]
                if cx is not None:
                    t = cx
            else:
                t = self.tm[f.context]
            if 'value' in x:
                process.add_exchange(f, d, value=x['value'], termination=t)

            if 'valueDict' in x:
                for k, val in x['valueDict'].items():
                    drr, fuu = k.split(':')
                    try:
                        rx = ref_x[fuu]
                    except KeyError:
                        process.show()
                        print('key: %s' % k)
                        print('flow: %s' % self[fuu])
                        raise
                    # assert rx.direction == drr
                    process.add_exchange(f, d, reference=rx, value=val, termination=t)

        if a_b_q is not None:
            alloc_q = self[a_b_q['externalId']]  # allocation quantity must be locally present
            if alloc_q is not None:
                try:
                    process.allocate_by_quantity(alloc_q)
                except ZeroAllocation:
                    pass

        return process

    def _make_entity(self, e, etype, ext_ref):
        if etype == 'process':
            return self._process_from_json(e, ext_ref)
        return super(LcArchive, self)._make_entity(e, etype, ext_ref)

    def entities_by_type(self, entity_type):
        if entity_type not in self._entity_types:
            entity_type = {
                'p': 'process',
                'f': 'flow',
                'q': 'quantity'
            }[entity_type[0]]
        return super(LcArchive, self).entities_by_type(entity_type)

    def serialize(self, exchanges=False, characterizations=False, values=False, domesticate=False):
        """

        :param exchanges:
        :param characterizations:
        :param values:
        :param domesticate: [False] if True, omit entities' origins so that they will appear to be from the new archive
         upon serialization
        :return:
        """
        j = super(LcArchive, self).serialize(characterizations=characterizations, values=values,
                                             domesticate=domesticate)
        j['processes'] = sorted([p.serialize(exchanges=exchanges, values=values,
                                             domesticate=domesticate, drop_fields=self._drop_fields['process'])
                                 for p in self.entities_by_type('process')],
                                key=lambda x: x['externalId'])
        if self._descendant:
            j['dataSourceType'] = 'LcArchive'  # re-instantiate as base class
        return j

    def _serialize_all(self, **kwargs):
        return self.serialize(exchanges=True, characterizations=True, values=True, **kwargs)



"""
This may be a bit of a protocol violation, but the role of the unittest framework here is to construct a persistent
catalog object.  The clever bit is that the object will be constructed entirely through unit test execution.

The catalog resources are specified in data_sources.local.  This catalog is used by computation unit-testing routines
to store persistent data sets to test against.

Because of naming conventions, this test file should run first, creating all the resources in the CATALOG_ROOT, which
again is a _persistent_ catalog (as opposed to TEST_ROOT, which is made to be blown away).  The later test files will
use the created resources to test functionality.

see ..local.py for resource configuration details.
"""

import unittest
import os
import json

from ..local import RESOURCES_CONFIG
from ...catalog import LcCatalog
from ...catalog_query import zap_inventory

from ...catalog.catalog_root import CATALOG_ROOT

resource_dir = os.path.join(CATALOG_ROOT, 'resources')


def _check_resource(ref):
    return os.path.exists(os.path.join(resource_dir, ref))


cat = LcCatalog(CATALOG_ROOT)


class LocalCatalog(unittest.TestCase):
    """
    Note hack of alphabetical naming of tests to force order
    """
    @classmethod
    def setUpClass(cls):
        """
        The setUpClass function creates the catalog and unpacks all test-enabled config specifications
        :return:
        """
        cls._configs = dict()
        cls._cat = cat
        for k, d in RESOURCES_CONFIG.items():
            if d.pop('enable_test', False):
                print('data source %s enabled' % k)
                obj = d.pop('source')
                cls._configs[k] = obj(**d)

    def test_folders(self):
        """
        Catalog creation should lead to the creation of the catalog root, subdirectories, and reference quantities.
        :return:
        """
        self.assertTrue(os.path.isdir(CATALOG_ROOT))
        self.assertTrue(os.path.isdir(resource_dir))
        self.assertTrue(os.path.exists(os.path.join(CATALOG_ROOT, 'reference-quantities.json')))

    def _check_reference(self, ref):
        return ref in self._cat.references

    def _check_interface(self, ref, iface):
        return ':'.join([ref, iface]) in self._cat.interfaces

    def test_a_make_resources(self):
        """
        Installs the resources from the config specifications, but only if the resources do not already exist.
        Something that's needed here is a way to update resources if the DataSource subclasses change-- presently the
        only way to do that is to delete the resources (and also any derived source files like indexes or backgrounds)
        This should simply run without errors, resulting in the creation of any missing resources
        :return:
        """
        # k is resource signifier, s is DataSource subclass
        for k, s in self._configs.items():
            for ref in s.references:
                if not self._check_reference(ref):
                    for res in s.make_resources(ref):
                        self._cat.add_resource(res)

    def test_b_number_of_resources(self):
        """
        This function checks that the number of resources created is correct
        :return:
        """
        # k is resource signifier, s is DataSource subclass
        for k, s in self._configs.items():
            for ref in s.references:
                nres = len([i for i in s.make_resources(ref)])
                with open(os.path.join(resource_dir, ref), 'r') as fp:
                    xres = len(json.load(fp)[ref])
                self.assertEqual(nres, xres, ref)

    def test_c_instantiate_ifaces(self):
        """
        This confirms that each interface specified can be created
        :return:
        """
        for k, s in self._configs.items():
            for ref in s.references:
                for iface in s.interfaces(ref):
                    iface = zap_inventory(iface, warn=False)
                    res = self._cat.get_resource(ref, iface, strict=True)
                    print(res)
                    res.check(self._cat)


if __name__ == '__main__':
    unittest.main()
